from typing import Optional

import numpy as np
from pydantic import BaseModel, validator, root_validator
from sklearn.utils.multiclass import unique_labels
from sklearn.multiclass import check_classification_targets
from sklearn.utils.validation import NotFittedError

__all__ = ['Array']

def _check_classes(y, classes=None):
    check_classification_targets(y)
    unique_y = unique_labels(y)

    classes_ = None
    if classes is None:
        classes_ = unique_y
    else:
        classes_np = np.array(classes).flatten()
        if not np.issubdtype(classes_np.dtype, np.integer):
            raise TypeError(f'Expected classes to be integers')

        _, idx = np.unique(classes_np, return_index=True)
        classes_ = classes_np[np.sort(idx)]
        unseen_labels = set(unique_y) - set(classes_np)
        if len(unseen_labels) > 0:
            raise ValueError(f'Encountered label(s) in `y` not present in specified classes - {unseen_labels}')

    return classes_

def _check_is_fitted(estimator, attributes=None, return_=False):
    fitted = False
    if attributes is None:
        fitted = any(attr.endswith('_') for attr in estimator.__dict__.keys() if '__' not in attr)
    else:
        fitted = all(hasattr(estimator, attr) for attr in attributes)

    if return_:
        return fitted

    if not fitted:
        raise NotFittedError(
            f'This {type(estimator).__name__} instance is not fitted yet. '
            "Call 'fit' with appropriate arguments before using this method."
        )

class _Validator(BaseModel):
    class Config:
        arbitrary_types_allowed = True

    @classmethod
    def fields(cls):
        return list(cls.__fields__.keys())

class _TypedArray(np.ndarray):
    @classmethod
    def __get_validators__(cls):
        yield cls.validate_type

    @classmethod
    def validate_type(cls, val):
        return np.array(val, dtype=cls.inner_type)

class _ArrayMeta(type):
    def __getitem__(self, t):
        return type('Array', (_TypedArray,), {'inner_type': t})

class Array(np.ndarray, metaclass=_ArrayMeta):
    pass

class _BaseSequenceValidator(_Validator):
    X: Array[None]
    lengths: Optional[Array[int]] = None
    y: Optional[Array[None]] = None

    @validator('X')
    def check_X(cls, X):
        X = np.atleast_1d(X)

        dim = X.ndim
        if dim == 1:
            X = X.reshape(-1, 1)
        elif dim > 2:
            raise ValueError('Expected `X` to have a maximum of two dimensions')

        if len(X) == 0:
            raise ValueError('Expected `X` to have at least one observation')

        return X

    @validator('lengths')
    def check_lengths(cls, value, values):
        X = values.get('X')
        if X is not None:
            len_X = len(X)

            # Treat whole input as one sequence if no lengths given (and try convert to numpy)
            lengths = np.array(len_X if value is None else value).flatten()
            total_lengths = lengths.sum()

            if total_lengths != len_X:
                raise ValueError(
                    'Sum of provided `lengths` does not match the length of `X` '
                    f'(sum(lengths)={total_lengths}, len(X)={len_X})'
                )
        else:
            raise ValueError('Unable to validate `lengths` as it depends on `X`')
        return lengths

    # Needs lengths
    @validator('y')
    def check_y(cls, value, values):
        y = value
        if y is not None:
            lengths = values.get('lengths')
            if lengths is not None:
                y = np.array(y).flatten()

                len_y = len(y)
                n_seqs = len(lengths)

                if len_y != n_seqs:
                    raise ValueError(
                        'Expected `y` to have the same number of elements as `lengths` '
                        f'(len(y)={len_y}, len(lengths)={n_seqs})'
                    )
            else:
                raise ValueError('Unable to validate `y` as it depends on `lengths`')
        return y

class _BaseSingleUnivariateSequenceValidator(_Validator):
    sequence: Array

    @validator('sequence')
    def check_sequence(cls, sequence):
        sequence = np.atleast_1d(sequence)

        dim = sequence.ndim
        if dim == 1:
            sequence = sequence.reshape(-1, 1)
        elif dim > 2:
            raise ValueError('Expected the sequence to have a maximum of two dimensions')

        if sequence.shape[1] > 1:
            raise ValueError('Expected the sequence to be univariate')

        if len(sequence) == 0:
            raise ValueError('Expected the sequence to have at least one observation')

        return sequence

class _BaseSingleMultivariateSequenceValidator(_Validator):
    sequence: Array

    @validator('sequence')
    def check_sequence(cls, sequence):
        sequence = np.atleast_1d(sequence)

        dim = sequence.ndim
        if dim == 1:
            sequence = sequence.reshape(-1, 1)
        elif dim > 2:
            raise ValueError('Expected sequence to have a maximum of two dimensions')

        if len(sequence) == 0:
            raise ValueError('Expected sequence to have at least one observation')

        return sequence

class _BaseUnivariateCategoricalSequenceValidator(_BaseSequenceValidator):
    X: Array[int]
    lengths: Optional[Array[int]] = None
    y: Optional[Array[None]] = None

class _UnivariateCategoricalSequenceClassifierValidator(_BaseSequenceValidator):
    X: Array[int]
    lengths: Optional[Array[int]] = None
    y: Array[int]

    @root_validator
    def check_X_is_1d(cls, values):
        X = values['X']
        if X.shape[1] > 1:
            raise ValueError('Only univariate categorical sequences are currently supported')
        return values

class _SingleUnivariateCategoricalSequenceValidator(_BaseSingleUnivariateSequenceValidator):
    sequence: Array[int]

class _BaseMultivariateFloatSequenceValidator(_BaseSequenceValidator):
    X: Array[np.float64]
    lengths: Optional[Array[int]] = None
    y: Optional[Array[None]] = None

class _MultivariateFloatSequenceClassifierValidator(_BaseSequenceValidator):
    X: Array[np.float64]
    lengths: Optional[Array[int]] = None
    y: Array[int]

class _MultivariateFloatSequenceRegressorValidator(_BaseSequenceValidator):
    X: Array[np.float64]
    lengths: Optional[Array[int]] = None
    y: Array[np.float64]

class _SingleUnivariateFloatSequenceValidator(_BaseSingleUnivariateSequenceValidator):
    sequence: Array[np.float64]

class _SingleMultivariateFloatSequenceValidator(_BaseSingleMultivariateSequenceValidator):
    sequence: Array[np.float64]



import pytest

import hmmlearn
import numpy as np

from sequentia.models import GaussianMixtureHMM
from sequentia.models.hmm.topologies import _topologies
from sequentia.datasets import load_digits
from sequentia.utils.validation import _check_is_fitted

from .....support.assertions import assert_equal, assert_not_equal
from .....support.itertools import combinations


@pytest.fixture(scope='module')
def random_state():
    return np.random.RandomState(0)


@pytest.fixture(scope='module')
def data(random_state):
    data_= load_digits(digits=[0])
    _, subset = data_.split(test_size=0.2, random_state=random_state, stratify=True)
    return subset


@pytest.fixture(scope='module')
def topology(request):
    return _topologies[request.param]


def assert_fit(hmm, data):
    assert hmm.n_seqs_ == len(data.lengths)
    assert (hmm.topology_ is not None) == (hmm.topology is not None)
    assert isinstance(hmm.model, hmmlearn.hmm.GMMHMM)
    assert len(hmm.model.monitor_.history) > 0
    assert _check_is_fitted(hmm, return_=True)


def test_gaussian_mixture_fit_n_states(data, random_state):
    hmm = GaussianMixtureHMM(n_states=7, random_state=random_state)

    assert_fit(hmm.fit(*data.X_lengths), data)

    assert hmm.n_states == 7

    assert hmm.model.startprob_.shape == (hmm.n_states,)
    assert hmm.model.transmat_.shape == (hmm.n_states, hmm.n_states)


def test_gaussian_mixture_fit_n_components(data, random_state):
    hmm = GaussianMixtureHMM(n_components=2, random_state=random_state)

    assert_fit(hmm.fit(*data.X_lengths), data)

    assert hmm.n_components == 2

    assert hmm.model.startprob_.shape == (hmm.n_states,)
    assert hmm.model.transmat_.shape == (hmm.n_states, hmm.n_states)

    n_features = data.X.shape[1]

    assert hmm.model.means_.shape == (hmm.n_states, hmm.n_components, n_features)
    assert hmm.model.covars_.shape == (hmm.n_states, hmm.n_components)
    assert hmm.model.weights_.shape == (hmm.n_states, hmm.n_components)


@pytest.mark.parametrize('covariance_type', ["spherical", "diag", "full", "tied"])
def test_gaussian_mixture_fit_covariance_type(data, random_state, covariance_type):
    hmm = GaussianMixtureHMM(covariance_type=covariance_type, random_state=random_state)

    assert_fit(hmm.fit(*data.X_lengths), data)

    assert hmm.covariance_type == covariance_type

    assert hmm.model.startprob_.shape == (hmm.n_states,)
    assert hmm.model.transmat_.shape == (hmm.n_states, hmm.n_states)

    n_features = data.X.shape[1]

    assert hmm.model.means_.shape == (hmm.n_states, hmm.n_components, n_features)
    assert hmm.model.weights_.shape == (hmm.n_states, hmm.n_components)

    if covariance_type == "spherical":
        assert hmm.model.covars_.shape == (hmm.n_states, hmm.n_components)
    elif covariance_type == "diag":
        assert hmm.model.covars_.shape == (hmm.n_states, hmm.n_components, n_features)
    elif covariance_type == "full":
        assert hmm.model.covars_.shape == (hmm.n_states, hmm.n_components, n_features, n_features)
    elif covariance_type == "tied":
        assert hmm.model.covars_.shape == (hmm.n_states, n_features, n_features)


def test_gaussian_mixture_fit_no_topology(data, random_state):
    hmm = GaussianMixtureHMM(topology=None, random_state=random_state)

    assert_fit(hmm.fit(*data.X_lengths), data)

    assert hmm.topology is None
    assert hmm.topology_ is None

    assert set(hmm.model.init_params) == set('stmcw')
    assert set(hmm.model.params) == set('stmcw')

    assert not hasattr(hmm, '_startprob')
    assert not hasattr(hmm, '_transmat')


@pytest.mark.parametrize('topology', ['ergodic', 'left-right', 'linear'], indirect=True)
@pytest.mark.parametrize('start_probs_type', ['uniform', 'random', None]) # None = custom
def test_gaussian_mixture_fit_set_start_probs(data, random_state, topology, start_probs_type):
    hmm = GaussianMixtureHMM(topology=topology.name, random_state=random_state)
    hmm.set_start_probs(start_probs_type or topology(hmm.n_states, random_state).random_start_probs())

    assert hmm.topology == topology.name
    if start_probs_type is not None:
        assert hmm._startprob == start_probs_type

    assert_fit(hmm.fit(*data.X_lengths), data)

    assert isinstance(hmm.topology_, topology)

    assert set(hmm.model.init_params) == set('mcw')
    assert set(hmm.model.params) == set('stmcw')

    hmm.topology_.check_start_probs(hmm._startprob) # transition matrix before fit
    hmm.topology_.check_start_probs(hmm.model.startprob_) # transition matrix after fit

    if start_probs_type == 'uniform':
        init_startprob = hmm.topology_.uniform_start_probs()
        assert_equal(hmm._startprob, init_startprob) # initial state probabilities should be uniform

    assert_not_equal(hmm._startprob, hmm.model.startprob_) # should update probabilities
    # assert_equal(hmm._startprob == 0, hmm.model.startprob_ == 0) # but locations of zeros (if any) shouldn't change


@pytest.mark.parametrize('topology', ['ergodic', 'left-right', 'linear'], indirect=True)
@pytest.mark.parametrize('transition_type', ['uniform', 'random', None]) # None = custom
def test_gaussian_mixture_fit_set_transitions(data, random_state, topology, transition_type):
    hmm = GaussianMixtureHMM(topology=topology.name, random_state=random_state)
    hmm.set_transitions(transition_type or topology(hmm.n_states, random_state).random_transitions())

    assert hmm.topology == topology.name
    if transition_type is not None:
        assert hmm._transmat == transition_type

    assert_fit(hmm.fit(*data.X_lengths), data)

    assert isinstance(hmm.topology_, topology)

    assert set(hmm.model.init_params) == set('mcw')
    assert set(hmm.model.params) == set('stmcw')

    hmm.topology_.check_transitions(hmm._transmat) # transition matrix before fit
    hmm.topology_.check_transitions(hmm.model.transmat_) # transition matrix after fit

    if transition_type == 'uniform':
        init_transmat = hmm.topology_.uniform_transitions()
        assert_equal(hmm._transmat, init_transmat) # transition probabilities should be uniform

    assert_not_equal(hmm._transmat, hmm.model.transmat_) # should update probabilities
    # assert_equal(hmm._transmat == 0, hmm.model.transmat_ == 0) # but locations of zeros (if any) shouldn't change


@pytest.mark.parametrize('freeze_params', combinations('stmcw'))
def test_gaussian_mixture_fit_freeze_unfreeze(data, random_state, freeze_params):
    hmm = GaussianMixtureHMM(topology='linear', n_components=2, n_states=2, random_state=random_state)
    hmm.freeze(freeze_params)

    assert_fit(hmm.fit(*data.X_lengths), data)

    assert set(hmm.model.params) == set('stmcw') - set(freeze_params)

    hmm.topology_.check_start_probs(hmm._startprob) # initial state dist. before fit
    hmm.topology_.check_start_probs(hmm.model.startprob_) # initial state dist. after fit
    assertion = assert_equal if 's' in freeze_params else assert_not_equal
    assertion(hmm._startprob, hmm.model.startprob_)

    hmm.topology_.check_transitions(hmm._transmat) # transition matrix before fit
    hmm.topology_.check_transitions(hmm.model.transmat_) # transition matrix after fit
    assertion = assert_equal if 't' in freeze_params else assert_not_equal
    assertion(hmm._transmat, hmm.model.transmat_)

    hmm.unfreeze(freeze_params)

    assert_fit(hmm.fit(*data.X_lengths), data)

    assert set(hmm.model.params) == set('stmcw')

    assert_not_equal(hmm._startprob, hmm.model.startprob_)
    assert_not_equal(hmm._transmat, hmm.model.transmat_)



import warnings

import numpy as np

class _Topology:
    """Represents a topology for a HMM, imposing restrictions on the transition matrix and initial state distribution.

    Parameters
    ----------
    n_states: int
        Number of states in the HMM.

    random_state: numpy.random.RandomState
        A random state object for reproducible randomness.
    """

    def __init__(self, n_states: int, random_state: np.random.RandomState):
        self.n_states = n_states
        self.random_state = random_state

    def uniform_start_probs(self) -> np.ndarray:
        """Sets the initial state distribution as a discrete uniform distribution.

        Returns
        -------
        initial: :class:`numpy:numpy.ndarray` (float)
            The initial state distribution of shape `(n_states,)`.
        """
        return np.ones(self.n_states) / self.n_states

    def random_start_probs(self) -> np.ndarray:
        """Sets the initial state distribution by randomly sampling probabilities generated by a Dirichlet distribution.

        Returns
        -------
        initial: :class:`numpy:numpy.ndarray` (float)
            The initial state distribution of shape `(n_states,)`.
        """
        return self.random_state.dirichlet(np.ones(self.n_states), size=1).flatten()

    def uniform_transitions(self) -> np.ndarray:
        """Sets the transition matrix as uniform (equal probability of transitioning
        to all other possible states from each state) corresponding to the topology.

        Returns
        -------
        transitions: :class:`numpy:numpy.ndarray` (float)
            The uniform transition matrix of shape `(n_states, n_states)`.
        """
        raise NotImplementedError

    def random_transitions(self) -> np.ndarray:
        """Sets the transition matrix as random (random probability of transitioning
        to all other possible states from each state) by sampling probabilities
        from a Dirichlet distribution - according to the topology.

        Returns
        -------
        transitions: :class:`numpy:numpy.ndarray` (float)
            The random transition matrix of shape `(n_states, n_states)`.
        """
        raise NotImplementedError

    def check_start_probs(self, initial: np.ndarray) -> None:
        """Validates an initial state distribution according to the topology's restrictions.

        Parameters
        ----------
        initial: numpy.ndarray (float)
            The initial state distribution to validate.
        """
        if not isinstance(initial, np.ndarray):
            raise TypeError('Initial state distribution must be a numpy.ndarray')
        if not initial.shape == (self.n_states,):
            raise ValueError('Initial state distribution must be of shape (n_states,)')
        if not np.isclose(initial.sum(), 1):
            raise ValueError('Initial state distribution must sum to one')
        return initial

    def check_transitions(self, transitions: np.ndarray) -> np.ndarray:
        """Validates a transition matrix according to the topology's restrictions.

        Parameters
        ----------
        transitions: numpy.ndarray (float)
            The transition matrix to validate.
        """
        if not isinstance(transitions, np.ndarray):
            raise TypeError('Transition matrix must be a numpy.ndarray')
        if not transitions.shape == (self.n_states, self.n_states):
            raise ValueError('Transition matrix must be of shape (n_states, n_states)')
        if not np.allclose(transitions.sum(axis=1), np.ones(self.n_states)):
            raise ValueError('Transition probabilities out of each state must sum to one')
        return transitions

class _ErgodicTopology(_Topology):
    """Represents the topology for an ergodic HMM, imposing non-zero probabilities in the transition matrix.

    Parameters
    ----------
    n_states: int
        Number of states in the HMM.

    random_state: numpy.random.RandomState
        A random state object for reproducible randomness.
    """

    name = "ergodic"

    def uniform_transitions(self) -> np.ndarray:
        """Sets the transition matrix as uniform (equal probability of transitioning
            to all other possible states from each state) corresponding to the topology.

        Returns
        -------
        transitions: :class:`numpy:numpy.ndarray` (float)
            The uniform transition matrix of shape `(n_states, n_states)`.
        """
        return np.ones((self.n_states, self.n_states)) / self.n_states

    def random_transitions(self) -> np.ndarray:
        """Sets the transition matrix as random (random probability of transitioning
        to all other possible states from each state) by sampling probabilities
        from a Dirichlet distribution - according to the topology.

        Returns
        -------
        transitions: :class:`numpy:numpy.ndarray` (float)
            The random transition matrix of shape `(n_states, n_states)`.
        """
        return self.random_state.dirichlet(np.ones(self.n_states), size=self.n_states)

    def check_transitions(self, transitions: np.ndarray) -> np.ndarray:
        """Validates a transition matrix according to the topology's restrictions.

        Parameters
        ----------
        transitions: numpy.ndarray (float)
            The transition matrix to validate.
        """
        super().check_transitions(transitions)
        if not np.all(transitions > 0):
            warnings.warn('Zero probabilities in ergodic transition matrix - these transition probabilities will not be learned')
        return transitions

class _LeftRightTopology(_Topology):
    """Represents the topology for a left-right HMM, imposing an upper-triangular transition matrix.

    Parameters
    ----------
    n_states: int
        Number of states in the HMM.

    random_state: numpy.random.RandomState
        A random state object for reproducible randomness.
    """

    name = "left-right"

    def uniform_transitions(self) -> np.ndarray:
        """Sets the transition matrix as uniform (equal probability of transitioning
            to all other possible states from each state) corresponding to the topology.

        Returns
        -------
        transitions: :class:`numpy:numpy.ndarray` (float)
            The uniform transition matrix of shape `(n_states, n_states)`.
        """
        upper_ones = np.triu(np.ones((self.n_states, self.n_states)))
        upper_divisors = np.triu(np.tile(np.arange(self.n_states, 0, -1), (self.n_states, 1)).T)
        lower_ones = np.tril(np.ones(self.n_states), k=-1)
        return upper_ones / (upper_divisors + lower_ones)

    def random_transitions(self) -> np.ndarray:
        """Sets the transition matrix as random (random probability of transitioning
        to all other possible states from each state) by sampling probabilities
        from a Dirichlet distribution, according to the topology.

        Returns
        -------
        transitions: :class:`numpy:numpy.ndarray` (float)
            The random transition matrix of shape `(n_states, n_states)`.
        """
        transitions = np.zeros((self.n_states, self.n_states))
        for i, row in enumerate(transitions):
            row[i:] = self.random_state.dirichlet(np.ones(self.n_states - i))
        return transitions

    def check_transitions(self, transitions: np.ndarray) -> np.ndarray:
        """Validates a transition matrix according to the topology's restrictions.

        Parameters
        ----------
        transitions: numpy.ndarray (float)
            The transition matrix to validate.
        """
        super().check_transitions(transitions)
        if not np.allclose(transitions, np.triu(transitions)):
            raise ValueError('Left-right transition matrix must be upper-triangular')
        return transitions

class _LinearTopology(_LeftRightTopology):
    """Represents the topology for a linear HMM.

    Parameters
    ----------
    n_states: int
        Number of states in the HMM.

    random_state: numpy.random.RandomState
        A random state object for reproducible randomness.
    """

    name = "linear"

    def uniform_transitions(self) -> np.ndarray:
        """Sets the transition matrix as uniform (equal probability of transitioning
            to all other possible states from each state) corresponding to the topology.

        Returns
        -------
        transitions: :class:`numpy:numpy.ndarray` (float)
            The uniform transition matrix of shape `(n_states, n_states)`.
        """
        transitions = np.zeros((self.n_states, self.n_states))
        for i, row in enumerate(transitions):
            size = min(2, self.n_states - i)
            row[i:(i + size)] = np.ones(size) / size
        return transitions

    def random_transitions(self) -> np.ndarray:
        """Sets the transition matrix as random (random probability of transitioning
        to all other possible states from each state) by sampling probabilities
        from a Dirichlet distribution, according to the topology.

        Returns
        -------
        transitions: :class:`numpy:numpy.ndarray` (float)
            The random transition matrix of shape `(n_states, n_states)`.
        """
        transitions = np.zeros((self.n_states, self.n_states))
        for i, row in enumerate(transitions):
            size = min(2, self.n_states - i)
            row[i:(i + size)] = self.random_state.dirichlet(np.ones(size))
        return transitions

    def check_transitions(self, transitions: np.ndarray) -> np.ndarray:
        """Validates a transition matrix according to the topology's restrictions.

        Parameters
        ----------
        transitions: numpy.ndarray (float)
            The transition matrix to validate.
        """
        super().check_transitions(transitions)
        if not np.allclose(transitions, np.diag(np.diag(transitions)) + np.diag(np.diag(transitions, k=1), k=1)):
            raise ValueError('Linear transition matrix must only consist of a diagonal and upper diagonal')
        return transitions

_topologies = {
    topology.name:topology
    for topology in (_ErgodicTopology, _LeftRightTopology, _LinearTopology)
}



from __future__ import annotations

import copy
import warnings
import pathlib
from typing import Optional, Tuple, Union, Iterator, IO

import numpy as np
from pydantic import NonNegativeInt, confloat
from sklearn.model_selection import train_test_split

from sequentia.utils.validation import _check_classes, _BaseSequenceValidator, Array

__all__ = ['SequentialDataset']


class SequentialDataset:
    """Utility wrapper for a generic sequential dataset."""

    def __init__(
        self,
        X: Array,
        y: Optional[Array] = None,
        lengths: Optional[Array[int]] = None,
        classes: Optional[Array[int]] = None
    ) -> SequentialDataset:
        """Initializes a :class:`.SequentialDataset`.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param y: Outputs corresponding to sequence(s) provided in ``X``.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :param classes: Set of possible class labels (only if ``y`` was provided with categorical values).

            - If not provided, these will be determined from the training data labels.
        """
        data = _BaseSequenceValidator(X=X, lengths=lengths, y=y)

        self._X = data.X
        self._y = data.y
        self._lengths = data.lengths

        self._classes = None
        if self._y is not None and np.issubdtype(self._y.dtype, np.integer):
            self._classes = _check_classes(self._y, classes)

        self._idxs = self._get_idxs(self.lengths)

        self._X_y = (self._X, self._y)
        self._X_lengths = (self._X, self._lengths)
        self._X_y_lengths = (self._X, self._y, self._lengths)


    def split(
        self,
        test_size: Optional[Union[NonNegativeInt, confloat(ge=0, le=1)]] = None,
        train_size: Optional[Union[NonNegativeInt, confloat(ge=0, le=1)]] = None,
        random_state: Optional[Union[NonNegativeInt, np.random.RandomState]] = None,
        shuffle: bool = True,
        stratify: bool = False
    ) -> Tuple[SequentialDataset, SequentialDataset]:
        """Splits the dataset into two partitions (train/test).

        See :func:`sklearn:sklearn.model_selection.train_test_split`.

        :param test_size: Size of the test partition.
        :param train_size: Size of the train partition.
        :param random_state: Seed or :class:`numpy:numpy.random.RandomState` object for reproducible pseudo-randomness.
        :param shuffle: Whether or not to shuffle the data before splitting. If ``shuffle=False`` then ``stratify`` must be ``False``.
        :param stratify: Whether or not to stratify the partitions by class labels.
        :return: Dataset partitions.
        """
        if stratify and self._y is None:
            warnings.warn('Cannot stratify with no provided outputs')
            stratify = None
        else:
            if stratify:
                if self._classes is None:
                    warnings.warn('Cannot stratify on non-categorical outputs')
                    stratify = None
                else:
                    stratify = self._y
            else:
                stratify = None

        idxs = np.arange(len(self._lengths))
        train_idxs, test_idxs = train_test_split(
            idxs,
            test_size=test_size,
            train_size=train_size,
            random_state=random_state,
            shuffle=shuffle,
            stratify=stratify
        )

        if self._y is None:
            X_train, y_train = self[train_idxs], None
            X_test, y_test = self[test_idxs], None
        else:
            X_train, y_train = self[train_idxs]
            X_test, y_test = self[test_idxs]

        lengths_train = self._lengths[train_idxs]
        lengths_test = self._lengths[test_idxs]
        classes = self._classes

        data_train = SequentialDataset(np.vstack(X_train), y=y_train, lengths=lengths_train, classes=classes)
        data_test = SequentialDataset(np.vstack(X_test), y=y_test, lengths=lengths_test, classes=classes)

        return data_train, data_test


    def iter_by_class(self) -> Iterator[Tuple[Array, Array, int]]:
        """Subsets the observation sequences by class.

        :raises: ``AttributeError`` - If ``y`` was not provided to :func:`__init__`, or is not categorical.
        :return: Generator iterating over classes, yielding:

            - ``X`` subset of sequences belonging to the class.
            - Lengths corresponding to the ``X`` subset.
            - Class used to subset ``X``.
        """
        if self._y is None:
            raise AttributeError('No `y` values were provided during initialization')

        if self._classes is None:
            raise RuntimeError('Cannot iterate by class on real-valued targets')

        for c in self._classes:
            ind = np.argwhere(self._y == c).flatten()
            X, _ = self[ind]
            lengths = self._lengths[ind]
            yield np.vstack(X), lengths, c


    @staticmethod
    def _get_idxs(lengths):
        ends = lengths.cumsum()
        starts = np.zeros_like(ends)
        starts[1:] = ends[:-1]
        return np.c_[starts, ends]


    @staticmethod
    def _iter_X(X, idxs):
        for start, end in idxs:
            yield X[start:end]


    def __len__(self):
        return len(self._lengths)


    def __getitem__(self, i):
        idxs = np.atleast_2d(self._idxs[i])
        X = [x for x in self._iter_X(self._X, idxs)]
        X = X[0] if isinstance(i, int) and len(X) == 1 else X
        return X if self._y is None else (X, self._y[i])


    def __iter__(self):
        for i in range(len(self)):
            yield self[i]


    @property
    def X(self) -> Array:
        """Observation sequences."""
        return self._X


    @property
    def y(self) -> Array:
        """Outputs corresponding to ``X``.

        :raises: ``AttributeError`` - If ``y`` was not provided to :func:`__init__`.
        """
        if self._y is None:
            raise AttributeError('No `y` values were provided during initialization')
        return self._y


    @property
    def lengths(self) -> Array[int]:
        """Lengths corresponding to ``X``."""
        return self._lengths


    @property
    def classes(self) -> Optional[Array[int]]:
        """Set of unique classes in ``y``. If ``y`` is not categorical, then ``None``."""
        return self._classes


    @property
    def idxs(self) -> Array[int]:
        """Observation sequence start and end indices."""
        return self._idxs


    @property
    def X_y(self) -> Tuple[Array, Array]:
        """Observation sequences and corresponding outputs.

        :raises: ``AttributeError`` - If ``y`` was not provided to :func:`__init__`.
        """
        if self._y is None:
            raise AttributeError('No `y` values were provided during initialization')
        return self._X_y


    @property
    def X_lengths(self) -> Tuple[Array, Array[int]]:
        """Observation sequences and corresponding lengths."""
        return self._X_lengths


    @property
    def X_y_lengths(self) -> Tuple[Array, Array, Array[int]]:
        """Observation sequences and corresponding outputs and lengths.

        :raises: ``AttributeError`` - If ``y`` was not provided to :func:`__init__`.
        """
        if self._y is None:
            raise AttributeError('No `y` values were provided during initialization')
        return self._X_y_lengths


    def save(self, path: Union[str, pathlib.Path, IO], compress: bool = True):
        """Stores the dataset in ``.npz`` format.

        See :func:`numpy:numpy.savez` and :func:`numpy:numpy.savez_compressed`.

        :param path: Location to store the dataset.
        :param compress: Whether or not to compress the dataset.

        See Also
        --------
        load:
            Loads a stored dataset in ``.npz`` format.
        """
        arrs = {
            'X': self._X,
            'lengths': self._lengths
        }

        if self._y is not None:
            arrs['y'] = self._y

        if self._classes is not None:
            arrs['classes'] = self._classes

        save_fun = np.savez_compressed if compress else np.savez
        save_fun(path, **arrs)


    @classmethod
    def load(cls, path: Union[str, pathlib.Path, IO]) -> SequentialDataset:
        """Loads a stored dataset in ``.npz`` format.

        See :func:`numpy:numpy.load`.

        :param path: Location to store the dataset.
        :return: The loaded dataset.

        See Also
        --------
        save:
            Stores the dataset in ``.npz`` format.
        """
        return cls(**np.load(path))


    def copy(self) -> SequentialDataset:
        """Creates a copy of the dataset.

        :return: Dataset copy.
        """
        params = {
            "X": copy.deepcopy(self._X),
            "lengths": copy.deepcopy(self._lengths),
        }

        if self._y is not None:
            params["y"] = copy.deepcopy(self._y)

        if self._classes is not None:
            params["classes"] = copy.deepcopy(self._classes)

        return SequentialDataset(**params)



from __future__ import annotations

import types
import joblib
import marshal
import warnings
import pathlib
from types import SimpleNamespace
from typing import Optional, Union, Callable, Tuple, List, Any, IO
from joblib import Parallel, delayed

import numpy as np
from pydantic import NegativeInt, NonNegativeInt, PositiveInt, confloat, validator
from dtaidistance import dtw, dtw_ndim
from sklearn.utils import check_random_state

from sequentia.utils.data import SequentialDataset
from sequentia.utils.multiprocessing import _effective_n_jobs
from sequentia.utils.decorators import (
    _validate_params,
    _override_params,
    _requires_fit,
    _check_plotting_dependencies,
)
from sequentia.utils.validation import (
    Array,
    _Validator,
    _BaseMultivariateFloatSequenceValidator,
    _SingleUnivariateFloatSequenceValidator,
    _SingleMultivariateFloatSequenceValidator,
)

dtw_cc = None
try:
    from dtaidistance import dtw_cc
except ImportError:
    pass

_defaults = SimpleNamespace(
    k=1,
    weighting=None,
    window=1,
    independent=False,
    use_c=False,
    n_jobs=1,
    random_state=None,
)


class _KNNValidator(_Validator):
    k: PositiveInt = _defaults.k
    weighting: Optional[Callable] = _defaults.weighting
    window: confloat(ge=0, le=1) = _defaults.window
    independent: bool = _defaults.independent
    use_c: bool = _defaults.use_c
    n_jobs: Union[NegativeInt, PositiveInt] = _defaults.n_jobs
    random_state: Optional[Union[NonNegativeInt, np.random.RandomState]] = _defaults.random_state


    @validator('use_c')
    def check_use_c(cls, value):
        use_c = value
        if use_c and dtw_cc is None:
            warnings.warn('DTAIDistance C library not available - using Python implementation', ImportWarning)
            use_c = False
        return use_c


    @validator('random_state')
    def check_random_state(cls, value):
        return check_random_state(value)


class _KNNMixin:
    _defaults = _defaults


    @_requires_fit
    @_override_params(['k', 'window', 'independent'])
    @_validate_params(using=_KNNValidator)
    def query_neighbors(
        self,
        X: Array[float],
        lengths: Optional[Array[int]] = None,
        sort: bool = True,
        **kwargs,
    ) -> Tuple[
        Array[int],
        Array[float],
        Array
    ]:
        """Queries the k-nearest training observation sequences to each sequence in ``X``.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :param sort: Whether to sort the neighbors in order of nearest to furthest.

        :param \*\*kwargs: Model parameters to temporarily override (*for experimentation purposes*).

            - ``k``: See :func:`__init__`.
            - ``window``: See :func:`__init__`.
            - ``independent``: See :func:`__init__`.

        :return: K-nearest neighbors for each sequence in ``X``.

            - Indices of the k-nearest training sequences.
            - DTW distances of the k-nearest training sequences.
            - Corresponding outputs of the k-nearest training sequences.
        """
        distances = self.compute_distance_matrix(X, lengths)
        partition_by = range(self.k) if sort else self.k
        k_idxs = np.argpartition(distances, partition_by, axis=1)[:, :self.k]
        k_distances = np.take_along_axis(distances, k_idxs, axis=1)
        k_outputs = self.y_[k_idxs]
        return k_idxs, k_distances, k_outputs


    @_requires_fit
    @_override_params(['window', 'independent'])
    @_validate_params(using=_KNNValidator)
    def compute_distance_matrix(
        self,
        X: Array[float],
        lengths: Optional[Array[int]] = None,
        **kwargs
    ) -> Array[float]:
        """Calculates a matrix of DTW distances between the sequences in ``X`` and the training sequences.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :param \*\*kwargs: Model parameters to temporarily override (*for experimentation purposes*).

            - ``window``: See :func:`__init__`.
            - ``independent``: See :func:`__init__`.

        :note: This method requires a trained classifier — see :func:`fit`.

        :return: DTW distance matrix.
        """
        data = _BaseMultivariateFloatSequenceValidator(X=X, lengths=lengths)

        n_jobs = _effective_n_jobs(self.n_jobs, data.lengths)
        dtw_ = self._dtw()

        row_chunk_idxs = np.array_split(SequentialDataset._get_idxs(data.lengths), n_jobs)
        col_chunk_idxs = np.array_split(self.idxs_, n_jobs)

        return np.vstack(
            Parallel(n_jobs=n_jobs, max_nbytes=None)(
                delayed(self._distance_matrix_row_chunk)(
                    row_idxs, col_chunk_idxs, data.X, n_jobs, dtw_
                ) for row_idxs in row_chunk_idxs
            )
        )


    @_override_params(['window', 'independent'])
    @_validate_params(using=_KNNValidator)
    def dtw(self, A: Array[float], B: Array[float], **kwargs) -> float:
        """Calculates the DTW distance between two univariate or multivariate sequences.

        :param A: The first sequence.

        :param B: The second sequence.

        :param \*\*kwargs: Model parameters to temporarily override (*for experimentation purposes*).

            - ``window``: See :func:`__init__`.
            - ``independent``: See :func:`__init__`.

        :return: DTW distance.
        """
        A = _SingleMultivariateFloatSequenceValidator(sequence=A).sequence
        B = _SingleMultivariateFloatSequenceValidator(sequence=B).sequence
        return self._dtw(A, B)


    @_check_plotting_dependencies
    @_override_params(['window'])
    @_validate_params(using=_KNNValidator)
    def plot_warping_path_1d(
        self,
        a: Array[float],
        b: Array[float],
        **kwargs
    ) -> 'matplotlib.axes.Axes':
        """Calculates the DTW matrix between two sequences and plots the optimal warping path.

        :param a: The first sequence.

        :param b: The second sequence.

        :note: Only supports univariate sequences.

        :param \*\*kwargs: Model parameters to temporarily override (*for experimentation purposes*).

            - ``window``: See :func:`__init__`.

        :return: Plot axes.
        """
        from dtaidistance import dtw_visualisation

        a = _SingleUnivariateFloatSequenceValidator(sequence=a).sequence
        b = _SingleUnivariateFloatSequenceValidator(sequence=b).sequence

        window = self._window(a, b)
        _, paths = dtw.warping_paths(a, b, window=window)
        best_path = dtw.best_path(paths)

        return dtw_visualisation.plot_warpingpaths(a, b, paths, best_path)


    @_check_plotting_dependencies
    @_requires_fit
    @_override_params(['window', 'independent'])
    @_validate_params(using=_KNNValidator)
    def plot_dtw_histogram(
        self,
        X: Array[float],
        lengths: Optional[Array[int]] = None,
        ax: Optional['matplotlib.axes.Axes'] = None,
        **kwargs
    ) -> 'matplotlib.axes.Axes':
        """Calculates DTW distances between ``X`` and training sequences, and plots a distance histogram.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :param ax: Plot axes. If ``None``, new axes are created.

        :param \*\*kwargs: Model parameters to temporarily override (*for experimentation purposes*).

            - ``window``: See :func:`__init__`.
            - ``independent``: See :func:`__init__`.

        :note: This method requires a trained classifier — see :func:`fit`.

        :return: Plot axes.
        """
        import matplotlib.pyplot as plt

        distances = self.compute_distance_matrix(X, lengths)

        if ax is None:
            _, ax = plt.subplots()
        ax.hist(distances.flatten())
        return ax


    @_check_plotting_dependencies
    @_requires_fit
    @_override_params(['weighting', 'window', 'independent'])
    @_validate_params(using=_KNNValidator)
    def plot_weight_histogram(
        self,
        X: Array[float],
        lengths: Optional[Array[int]] = None,
        ax: Optional['matplotlib.axes.Axes'] = None,
        **kwargs
    ) -> 'matplotlib.axes.Axes':
        """Calculates DTW weights between ``X`` and training sequences, and plots a weight histogram.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :param ax: Plot axes. If ``None``, new axes are created.

        :param \*\*kwargs: Model parameters to temporarily override (*for experimentation purposes*).

            - ``weighting``: See :func:`__init__`.
            - ``window``: See :func:`__init__`.
            - ``independent``: See :func:`__init__`.

        :note: This method requires a trained classifier — see :func:`fit`.

        :return: Plot axes.
        """
        import matplotlib.pyplot as plt

        distances = self.compute_distance_matrix(X, lengths)
        weightings = self._weighting()(distances)

        if ax is None:
            _, ax = plt.subplots()
        ax.hist(weightings.flatten())
        return ax


    def _dtw1d(self, a: Array[float], b: Array[float], window: int) -> float:
        """Computes the DTW distance between two univariate sequences."""
        return dtw.distance(a, b, use_c=self.use_c, window=window)


    def _window(self, A: Array[float], B: Array[float]) -> int:
        """TODO"""
        return int(self.window * min(len(A), len(B)))


    def _dtwi(self, A: Array[float], B: Array[float]) -> float:
        """Computes the multivariate DTW distance as the sum of the pairwise per-feature DTW distances,
        allowing each feature to be warped independently."""
        window = self._window(A, B)
        return np.sum([self._dtw1d(A[:, i], B[:, i], window) for i in range(A.shape[1])])


    def _dtwd(self, A: Array[float], B: Array[float]) -> float:
        """Computes the multivariate DTW distance so that the warping of the features depends on each other,
        by modifying the local distance measure."""
        window = self._window(A, B)
        return dtw_ndim.distance(A, B, use_c=self.use_c, window=window)


    def _dtw(self) -> Callable:
        """TODO"""
        return self._dtwi if self.independent else self._dtwd


    def _weighting(self) -> Callable:
        """TODO"""
        return self.weighting if callable(self.weighting) else lambda x: np.ones_like(x)


    def _distance_matrix_row_chunk(
        self,
        row_idxs: Array[int],
        col_chunk_idxs: List[Array[int]],
        X: Array[float],
        n_jobs: int,
        dist: Callable
    ) -> Array[float]:
        """Calculates a distance sub-matrix for a subset of rows over all columns.

        TODO
        """
        return np.hstack(
            Parallel(n_jobs=n_jobs, max_nbytes=None)(
                delayed(self._distance_matrix_row_col_chunk)(
                    col_idxs, row_idxs, X, dist
                ) for col_idxs in col_chunk_idxs
            )
        )


    def _distance_matrix_row_col_chunk(
        self,
        col_idxs: Array[int],
        row_idxs: Array[int],
        X: Array[float],
        dist: Callable
    ) -> Array[float]:
        """Calculates a distance sub-matrix for a subset of rows and columns.

        TODO
        """
        distances = np.zeros((len(row_idxs), len(col_idxs)))
        for i, x_row in enumerate(SequentialDataset._iter_X(X, row_idxs)):
            for j, x_col in enumerate(SequentialDataset._iter_X(self.X_, col_idxs)):
                distances[i, j] = dist(x_row, x_col)
        return distances


    @_requires_fit
    def save(self, path: Union[str, pathlib.Path, IO]):
        """Serializes and saves a fitted KNN estimator.

        :param path: Location to save the serialized estimator.

        :note: This method requires a trained classifier — see :func:`fit`.

        See Also
        --------
        load:
            Loads and deserializes a fitted KNN estimator.
        """
        # Fetch main parameters and fitted values
        state = {
            'params': self.get_params(),
            'fitted': {k:v for k, v in self.__dict__.items() if k.endswith('_')}
        }

        # Serialize weighting function
        if self.weighting is None:
            state['params']['weighting'] = self.weighting
        else:
            state['params']['weighting'] = marshal.dumps(
                (self.weighting.__code__, self.weighting.__name__)
            )

        # Serialize model
        joblib.dump(state, path)


    @classmethod
    def load(cls, path: Union[str, pathlib.Path, IO]):
        """Loads and deserializes a fitted KNN estimator.

        :param path: Location to load the serialized estimator from.

        :return: Fitted KNN estimator.

        See Also
        --------
        save:
            Serializes and saves a fitted KNN estimator.
        """
        state = joblib.load(path)

        # Deserialize weighting function
        if state['params']['weighting'] is not None:
            weighting, name = marshal.loads(state['params']['weighting'])
            state['params']['weighting'] = types.FunctionType(weighting, globals(), name)

        # Set main parameters
        model = cls(**state['params'])

        # Set fitted values
        for k, v in state['fitted'].items():
            setattr(model, k, v)

        # Return deserialized model
        return model



from __future__ import annotations

from typing import Optional, Union, Callable, Any

import numpy as np
from pydantic import NegativeInt, NonNegativeInt, PositiveInt, confloat
from sklearn.utils import check_random_state

from sequentia.models.knn.base import _KNNMixin, _KNNValidator, _defaults
from sequentia.models.base import _Regressor

from sequentia.utils.decorators import _validate_params, _requires_fit, _override_params
from sequentia.utils.data import SequentialDataset
from sequentia.utils.validation import (
    Array,
    _MultivariateFloatSequenceRegressorValidator
)

__all__ = ['KNNRegressor']


class KNNRegressor(_KNNMixin, _Regressor):
    """A k-nearest neighbor regressor that uses DTW as a distance measure for sequence comparison.

    The regressor computes the output as a distance weighted average of the outputs of the sequences within the DTW k-neighborhood of the sequence being predicted.
    """

    @_validate_params(using=_KNNValidator)
    def __init__(
        self,
        *,
        k: PositiveInt = _defaults.k,
        weighting: Optional[Callable] = _defaults.weighting,
        window: confloat(ge=0, le=1) = _defaults.window,
        independent: bool = _defaults.independent,
        use_c: bool = _defaults.use_c,
        n_jobs: Union[NegativeInt, PositiveInt] = _defaults.n_jobs,
        random_state: Optional[Union[NonNegativeInt, np.random.RandomState]] = _defaults.random_state
    ) -> KNNRegressor:
        """Initializes the :class:`.KNNRegressor`.

        :param k: Number of neighbors.

        :param weighting: A callable that specifies how distance weighting should be performed.
            The callable should accept a :class:`numpy:numpy.ndarray` of DTW distances, apply an element-wise weighting transformation
            to the matrix of DTW distances, then return an equally-sized :class:`numpy:numpy.ndarray` of weightings.
            If ``None``, then a uniform weighting of 1 will be applied to all distances.

        :param window: The size of the Sakoe—Chiba band global constrant as a fraction of the length of the shortest of the two sequences being compared.

            - A larger window will give more freedom to the DTW alignment, allowing more deviation but leading to potentially slower computation.
              A window of 1 is equivalent to full DTW computation with no global constraint applied.
            - A smaller window will restrict the DTW alignment, and possibly speed up the DTW computation.
              A window of 0 is equivalent to Euclidean distance.

        :param independent: Whether or not to allow features to be warped independently from each other. See [#dtw_multi]_ for an overview of independent and dependent dynamic time warping.

        :param use_c: Whether or not to use fast pure C compiled functions from `dtaidistance <https://github.com/wannesm/dtaidistance>`__ to perform the DTW computations.

        :param n_jobs: Maximum number of concurrently running workers.

            - If 1, no parallelism is used at all (useful for debugging).
            - If -1, all CPUs are used.
            - If < -1, ``(n_cpus + 1 + n_jobs)`` are used — e.g. ``n_jobs=-2`` uses all but one.

        :param random_state: Seed or :class:`numpy:numpy.random.RandomState` object for reproducible pseudo-randomness.
        """
        #: Number of neighbors.
        self.k = k
        #: A callable that specifies how distance weighting should be performed.
        self.weighting = weighting
        #: The size of the Sakoe—Chiba band global constrant as a fraction of the length of the shortest of the two sequences being compared.
        self.window = window
        #: Whether or not to allow features to be warped independently from each other.
        self.independent = independent
        #: Set of possible class labels.
        self.use_c = use_c
        #: Maximum number of concurrently running workers.
        self.n_jobs = n_jobs
        #: Seed or :class:`numpy:numpy.random.RandomState` object for reproducible pseudo-randomness.
        self.random_state = random_state


    def fit(
        self,
        X: Array[float],
        y: Array[float],
        lengths: Optional[Array[int]] = None
    ) -> KNNRegressor:
        """Fits the regressor to the sequence(s) in ``X``.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param y: Outputs corresponding to sequence(s) provided in ``X``.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :return: The fitted regressor.
        """
        data = _MultivariateFloatSequenceRegressorValidator(X=X, y=y, lengths=lengths)
        self.X_ = data.X
        self.y_ = data.y
        self.lengths_ = data.lengths
        self.idxs_ = SequentialDataset._get_idxs(data.lengths)
        self.random_state_ = check_random_state(self.random_state)
        return self


    @_requires_fit
    def predict(
        self,
        X: Array[float],
        lengths: Optional[Array[int]] = None
    ) -> Array[float]:
        """Predicts outputs for the sequence(s) in ``X``.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :note: This method requires a trained regressor — see :func:`fit`.

        :return: Output predictions.
        """
        _, k_distances, k_outputs = self.query_neighbors(X, lengths, sort=False)
        k_weightings = self._weighting()(k_distances)
        return (k_outputs * k_weightings).sum(axis=1) / k_weightings.sum(axis=1)


    def fit_predict(
        self,
        X: Array[float],
        y: Array[float],
        lengths: Optional[Array[int]] = None
    ) -> Array[float]:
        """Fits the regressor to the sequence(s) in ``X`` and predicts outputs for ``X``.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param y: Outputs corresponding to sequence(s) provided in ``X``.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :return: Output predictions.
        """
        return super().fit_predict(X, y, lengths)


    @_requires_fit
    def score(
        self,
        X: Array[float],
        y: Array[float],
        lengths: Optional[Array[int]] = None,
        sample_weight: Optional[Any] = None
    ) -> float:
        """Calculates the coefficient of determination (R\ :sup:`2`) for the sequence(s) in ``X``.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param y: Outputs corresponding to the observation sequence(s) in ``X``.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :param sample_weight: See :func:`sklearn:sklearn.metrics.r2_score`.

        :note: This method requires a trained regressor — see :func:`fit`.

        :return: Coefficient of determination.
        """
        return super().score(X, y, lengths, sample_weight)


    @_validate_params(using=_KNNValidator)
    @_override_params(['k', 'weighting', 'window', 'independent', 'use_c', 'n_jobs'], temporary=False)
    def set_params(self, **kwargs):
        return self



import joblib
from typing import Optional

from sequentia.utils.validation import Array

def _effective_n_jobs(n_jobs: int, lengths: Optional[Array[int]] = None) -> int:
    n_jobs_ = 1
    if lengths is not None:
        n_jobs_ = min(joblib.effective_n_jobs(n_jobs), len(lengths))
    return n_jobs_



from typing import Optional

from sklearn.base import TransformerMixin, BaseEstimator

from sequentia.utils.validation import Array

__all__ = ['Transform']


class Transform(TransformerMixin, BaseEstimator):
    def fit_transform(
        self,
        X: Array,
        lengths: Optional[Array] = None,
    ) -> Array:
        """Fits the transformer to the sequence(s) in ``X`` and returns a transformed version of ``X``.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :return: The transformed data.
        """
        return self.fit(X, lengths).transform(X, lengths)



import pytest

import numpy as np
from pydantic import ValidationError

from sklearn.preprocessing import scale
from sklearn.decomposition import PCA
from sklearn.utils.validation import check_is_fitted
from sklearn.exceptions import NotFittedError
from sklearn.utils._param_validation import InvalidParameterError

from sequentia.datasets import load_digits
from sequentia.preprocessing import IndependentFunctionTransformer, mean_filter
from sequentia.pipeline import Pipeline
from sequentia.models import KNNClassifier

from sequentia.utils import SequentialDataset

from ..support.assertions import assert_equal, assert_not_equal


@pytest.fixture(scope='module')
def random_state(request):
    return np.random.RandomState(0)


@pytest.fixture(scope='module')
def data(random_state):
    data_= load_digits(digits=[0])
    _, subset = data_.split(test_size=0.05, random_state=random_state, stratify=True)
    return subset


def test_pipeline_with_transforms(data):
    # create pipeline with a stateless and stateful transform
    pipeline = Pipeline([
        ("scale", IndependentFunctionTransformer(
            scale,
            inverse_func=lambda x: x,
            check_inverse=False
        )),
        ("pca", PCA(n_components=1)),
    ])

    # check that transforming without fitting doesn't work
    with pytest.raises(NotFittedError):
        pipeline.transform(*data.X_lengths)

    # check that fitting without y works
    check_is_fitted(pipeline.fit(*data.X_lengths))
    for estimator in pipeline.named_steps.values():
        check_is_fitted(estimator)

    # check that fitting with y works
    check_is_fitted(pipeline.fit(*data.X_y_lengths))
    for estimator in pipeline.named_steps.values():
        check_is_fitted(estimator)

    # check that transforming after fit works
    Xt = pipeline.transform(*data.X_lengths)
    assert_not_equal(data.X, Xt)
    assert Xt.shape == (len(data.X), 1)

    # check that fit_transform works
    Xt = pipeline.fit_transform(*data.X_lengths)
    assert_not_equal(data.X, Xt)
    assert Xt.shape == (len(data.X), 1)

    # check that inverse_transform works
    Xi = pipeline.inverse_transform(Xt, data.lengths)
    assert_not_equal(Xt, Xi)

    # check that prediction functions relying on X and lengths don't work
    for func in ('predict', 'predict_proba'):
        with pytest.raises(AttributeError):
            getattr(pipeline, func)(*data.X_lengths)

    # check that fit_predict doesn't work
    with pytest.raises(AttributeError):
        pipeline.fit_predict(*data.X_y_lengths)

    # check that score works if the final transform implements it, with y
    pipeline.score(*data.X_y_lengths)

    # check that score works if the final transform implements it, without y
    pipeline.score(data.X, lengths=data.lengths)


def test_pipeline_with_estimator(data):
    pipeline = Pipeline([
        ("knn", KNNClassifier(k=1)),
    ])

    # check that transforming doesn't work
    with pytest.raises(AttributeError):
        pipeline.transform(*data.X_lengths)

    # check that fitting without y doesn't work
    with pytest.raises(ValidationError):
        pipeline.fit(data.X, lengths=data.lengths)

    # check that fitting with y works
    check_is_fitted(pipeline.fit(*data.X_y_lengths))
    for estimator in pipeline.named_steps.values():
        check_is_fitted(estimator)

    # check that transforming doesn't work
    with pytest.raises(AttributeError):
        pipeline.transform(*data.X_lengths)

    # check that fit_transform doesn't work
    with pytest.raises(AttributeError):
        pipeline.fit_transform(*data.X_lengths)

    # check that inverse_transform doesn't work
    with pytest.raises(AttributeError):
        pipeline.inverse_transform(*data.X_lengths)

    # check that predict works
    y_pred = pipeline.predict(*data.X_lengths)
    assert y_pred.shape == data.y.shape
    assert set(y_pred) == set(data.classes)

    # check that predict_proba works
    proba_pred = pipeline.predict_proba(*data.X_lengths)
    assert proba_pred.shape == (len(data), len(data.classes))

    # check that fit_predict works
    y_pred = pipeline.fit_predict(*data.X_y_lengths)
    # check that all steps are fitted
    check_is_fitted(pipeline.fit(*data.X_y_lengths))
    for estimator in pipeline.named_steps.values():
        check_is_fitted(estimator)
    # check that predictions are valid
    assert y_pred.shape == data.y.shape
    assert set(y_pred) == set(data.classes)

    # check that score with y works
    pipeline.score(*data.X_y_lengths)

    # check that score without y doesn't work
    with pytest.raises(InvalidParameterError):
        pipeline.score(data.X, lengths=data.lengths)


def test_pipeline_with_transforms_and_estimator(data):
    pipeline = Pipeline([
        ("scale", IndependentFunctionTransformer(
            scale,
            inverse_func=lambda x: x,
            check_inverse=False
        )),
        ("pca", PCA(n_components=1)),
        ("knn", KNNClassifier(k=1)),
    ])

    # check that transforming doesn't work
    with pytest.raises(AttributeError):
        pipeline.transform(*data.X_lengths)

    # check that fitting without y doesn't work
    with pytest.raises(ValidationError):
        pipeline.fit(data.X, lengths=data.lengths)

    # check that fitting with y works
    check_is_fitted(pipeline.fit(*data.X_y_lengths))
    for estimator in pipeline.named_steps.values():
        check_is_fitted(estimator)
    # check that X values were transformed
    assert_not_equal(data.X, pipeline[-1].X_)
    assert pipeline[-1].X_.shape == (len(data.X), 1)

    # check that transforming doesn't work
    with pytest.raises(AttributeError):
        pipeline.transform(*data.X_lengths)

    # check that fit_transform doesn't work
    with pytest.raises(AttributeError):
        pipeline.fit_transform(*data.X_lengths)

    # check that inverse_transform doesn't work
    with pytest.raises(AttributeError):
        pipeline.inverse_transform(*data.X_lengths)

    # check that predict works
    y_pred = pipeline.predict(*data.X_lengths)
    assert y_pred.shape == data.y.shape
    assert set(y_pred) == set(data.classes)

    # check that predict_proba works
    proba_pred = pipeline.predict_proba(*data.X_lengths)
    assert proba_pred.shape == (len(data), len(data.classes))

    # check that fit_predict works
    y_pred = pipeline.fit_predict(*data.X_y_lengths)
    # check that all steps are fitted
    check_is_fitted(pipeline.fit(*data.X_y_lengths))
    for estimator in pipeline.named_steps.values():
        check_is_fitted(estimator)
    # check that predictions are valid
    assert y_pred.shape == data.y.shape
    assert set(y_pred) == set(data.classes)
    # check that X values were transformed
    assert_not_equal(data.X, pipeline[-1].X_)
    assert pipeline[-1].X_.shape == (len(data.X), 1)

    # check that score with y works
    pipeline.score(*data.X_y_lengths)

    # check that score without y doesn't work
    with pytest.raises(InvalidParameterError):
        pipeline.score(data.X, lengths=data.lengths)



import os
import math
import pytest
from copy import deepcopy
from tempfile import TemporaryDirectory
from unittest.mock import Mock

import numpy as np

from sequentia.datasets import load_digits
from sequentia.models.knn import KNNClassifier
from sequentia.utils.validation import _check_is_fitted

from ....support.assertions import assert_equal

n_classes = 3


@pytest.fixture(scope='module')
def random_state(request):
    return np.random.RandomState(1)


@pytest.fixture(scope='module')
def dataset():
    return load_digits(digits=range(n_classes))


def assert_fit(clf, data):
    assert hasattr(clf, 'X_')
    assert hasattr(clf, 'y_')
    assert hasattr(clf, 'lengths_')
    assert hasattr(clf, 'idxs_')
    assert _check_is_fitted(clf, return_=True)
    assert_equal(clf.X_, data.X)
    assert_equal(clf.y_, data.y)
    assert_equal(clf.lengths_, data.lengths)


@pytest.mark.parametrize('k', [1, 2, 5])
@pytest.mark.parametrize('weighting', [None, lambda x: np.exp(-x)])
def test_classifier_e2e(request, k, weighting, dataset, random_state):
    clf = KNNClassifier(k=k, weighting=weighting, random_state=random_state)

    assert clf.k == k
    assert clf.weighting == weighting

    data = dataset.copy()
    data._X = data._X[:, :1] # only use one feature
    subset, _ = data.split(test_size=0.98, random_state=random_state, stratify=True)
    train, test = subset.split(test_size=0.2, random_state=random_state, stratify=True)

    assert_fit(clf.fit(*train.X_y_lengths), train)
    params = clf.get_params()

    scores_pred = clf.predict_scores(*test.X_lengths)
    assert scores_pred.shape == (len(test), n_classes)

    proba_pred = clf.predict_proba(*test.X_lengths)
    assert proba_pred.shape == (len(test), n_classes)
    assert_equal(proba_pred.sum(axis=1), 1)
    assert ((proba_pred >= 0) & (proba_pred <= 1)).all()

    y_pred = clf.predict(*test.X_lengths)
    assert np.issubdtype(y_pred.dtype, np.integer)
    assert y_pred.shape == (len(test),)
    assert set(y_pred).issubset(set(range(n_classes)))

    acc = clf.score(*test.X_y_lengths)
    assert 0 <= acc <= 1

    # check serialization/deserialization
    with TemporaryDirectory() as temp_dir:
        model_path = f"{temp_dir}/{request.node.originalname}.model"
        # check that save works
        clf.save(model_path)
        assert os.path.isfile(model_path)
        # check that load works
        clf = KNNClassifier.load(model_path)
        assert (set(clf.get_params().keys()) - set(['weighting'])) == (set(params.keys()) - set(['weighting']))
        # sanity check that custom weighting functions are the same
        if weighting:
            x = random_state.rand(100)
            assert_equal(weighting(x), clf.weighting(x))
        # check that loaded model is fitted and can make predictions
        assert_fit(clf, train)
        y_pred_load = clf.predict(*test.X_lengths)
        if k == 1:
            # predictions should be same as before
            assert_equal(y_pred, y_pred_load)


def test_classifier_predict_train(dataset, random_state):
    """Should be able to perfectly predict training data with k=1"""
    clf = KNNClassifier(k=1, random_state=random_state)

    data = dataset.copy()
    data._X = data._X[:, :1] # only use one feature
    train, _ = data.split(train_size=0.05, random_state=random_state, stratify=True)

    assert_fit(clf.fit(*train.X_y_lengths), train)
    assert math.isclose(clf.score(*train.X_y_lengths), 1.)


@pytest.mark.parametrize('classes', [[0, 1, 2], [2, 0, 1]])
def test_classifier_compute_scores(classes, random_state):
    clf = KNNClassifier(k=5)
    clf.random_state_ = random_state
    clf.classes_ = np.array(classes)

    labels = np.array([
        [0, 2, 1, 2, 2],
        [0, 1, 1, 2, 0],
        [1, 0, 0, 1, 2],
        [0, 0, 0, 1, 1]
    ])
    weightings = np.ones_like(labels)

    scores = clf._compute_scores(labels, weightings)
    if np.allclose(classes, [0, 1, 2]):
        assert_equal(scores, [
            [1, 1, 3],
            [2, 2, 1],
            [2, 2, 1],
            [3, 2, 0]
        ])
    elif np.allclose(classes, [2, 0, 1]):
        assert_equal(scores, [
            [3, 1, 1],
            [1, 2, 2],
            [1, 2, 2],
            [0, 3, 2]
        ])


@pytest.mark.parametrize('classes', [[0, 1, 2], [2, 0, 1]])
def test_classifier_find_max_labels_chunk(classes, random_state):
    clf = KNNClassifier()
    clf.random_state_ = random_state
    clf.classes_ = np.array(classes)

    score_chunk = np.array([
        [10, 20, 20],
        [10, 30, 20],
        [10, 10, 10],
        [10, 10, 20]
    ])

    max_labels = clf._find_max_labels_chunk(score_chunk)
    if np.allclose(classes, [0, 1, 2]):
        assert max_labels[0] in (1, 2)
        assert max_labels[1] == 1
        assert max_labels[2] in (0, 1, 2)
        assert max_labels[3] == 2
    elif np.allclose(classes, [2, 0, 1]):
        assert max_labels[0] in (0, 1)
        assert max_labels[1] == 0
        assert max_labels[2] in (0, 1, 2)
        assert max_labels[3] == 1


@pytest.mark.parametrize('tie', [True, False])
def test_classifier_multi_argmax(tie):
    if tie:
        arr = np.array([3, 2, 4, 1, 3, 4, 4, 0, 2, 4])
        assert_equal(
            KNNClassifier._multi_argmax(arr),
            np.array([2, 5, 6, 9])
        )
    else:
        arr = np.array([3, 2, 1, 1, 3, 4, 1, 0, 2, 0])
        assert_equal(
            KNNClassifier._multi_argmax(arr),
            np.array([5])
        )


@pytest.mark.parametrize('k', [1, 2, 5])
@pytest.mark.parametrize('sort', [True, False])
def test_classifier_query_neighbors(k, sort, dataset, random_state):
    clf = KNNClassifier(k=k, random_state=random_state)

    data = dataset.copy()
    data._X = data._X[:, :1] # only use one feature
    subset, _ = data.split(test_size=0.98, random_state=random_state, stratify=True)
    train, test = subset.split(test_size=0.2, random_state=random_state, stratify=True)

    assert_fit(clf.fit(*train.X_y_lengths), train)

    k_idxs, k_distances, k_labels = clf.query_neighbors(*test.X_lengths, sort=sort)

    # check that indices are between 0 and len(train)
    assert np.issubdtype(k_idxs.dtype, np.integer)
    assert k_idxs.shape == (len(test), clf.k)
    assert set(k_idxs.flatten()).issubset(set(np.arange(len(train))))

    # check that distances are sorted if sort=True
    np.issubdtype(k_distances.dtype, np.floating)
    assert k_distances.shape == (len(test), clf.k)
    if sort and k > 1:
        assert (k_distances[:, 1:] >= k_distances[:, :-1]).all()

    # check that labels are a subset of training labels + check that labels match indices
    assert np.issubdtype(k_labels.dtype, np.integer)
    assert k_labels.shape == (len(test), clf.k)
    assert set(k_labels.flatten()).issubset(set(train.y))
    assert_equal(train.y[k_idxs], k_labels)


def test_classifier_compute_distance_matrix(dataset, random_state):
    clf = KNNClassifier()

    data = dataset.copy()
    data._X = data._X[:, :1] # only use one feature
    subset, _ = data.split(test_size=0.98, random_state=random_state, stratify=True)
    train, test = subset.split(test_size=0.2, random_state=random_state, stratify=True)

    assert_fit(clf.fit(*train.X_y_lengths), train)

    distances = clf.compute_distance_matrix(*test.X_lengths)
    assert distances.shape == (len(test), len(train))


def test_classifier_distance_matrix_row_col_chunk():
    clf = KNNClassifier()

    clf.X_ = np.expand_dims(np.arange(7), axis=-1)
    col_idxs = np.array([[0, 1], [1, 3], [4, 7]]) # lengths = 1, 2, 3

    X = np.expand_dims(np.arange(14), axis=-1)
    row_idxs = np.array([[0, 2], [2, 5], [5, 9], [9, 14]]) # lengths = 2, 3, 4, 5

    distances = clf._distance_matrix_row_col_chunk(col_idxs, row_idxs, X, lambda x1, x2: len(x1) - len(x2))
    assert_equal(
        distances,
        np.array([
            [1, 0, -1],
            [2, 1,  0],
            [3, 2,  1],
            [4, 3,  2]
        ])
    )



"""
IndependentFunctionTransformer is an adapted version of FunctionTransformer
from the sklearn.preprocessing module, and largely relies on its source code.

Below is the original license from Scikit-Learn, copied on 31st December 2022
from https://github.com/scikit-learn/scikit-learn/blob/main/COPYING.

---

BSD 3-Clause License

Copyright (c) 2007-2022 The scikit-learn developers.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from __future__ import annotations

import warnings
from typing import Callable, Optional, Dict, Any

import numpy as np
from pydantic import PositiveInt
from sklearn.utils.validation import _allclose_dense_sparse, check_array
from scipy.signal import medfilt2d, convolve

from sequentia.preprocessing.base import Transform
from sequentia.utils.validation import _BaseSequenceValidator, Array
from sequentia.utils.data import SequentialDataset

__all__ = ["IndependentFunctionTransformer", "mean_filter", "median_filter"]


class IndependentFunctionTransformer(Transform):
    """Constructs a transformer from an arbitrary callable,
    applying the transform independently to each sequence.

    This transform forwards its ``X`` and ``lengths`` arguments
    to a user-defined function or function object and returns the result of this
    function. This is useful for stateless transformations such as taking the
    log of frequencies, doing custom scaling, etc.
    Note: If a lambda is used as the function, then the resulting
    transformer will not be pickleable.

    This works conveniently with functions in :mod:`sklearn.preprocessing`
    such as :func:`~sklearn.preprocessing.scale` or :func:`~sklearn.preprocessing.normalize`.

    :note: This is a stateless transform, meaning :func:`fit` and :func:`fit_transform` will not fit on any data.

    See Also
    --------
    :class:`sklearn.preprocessing.FunctionTransformer`
        :class:`.IndependentFunctionTransformer` is based on this class,
        which applies the callable to the entire input array ``X`` as if it were a single sequence.
        Read more in the :ref:`User Guide <function_transformer>`.

    Examples
    --------
    Using an :class:`IndependentFunctionTransformer` with :func:`sklearn.preprocessing.minmax_scale` to
    scale features to the range [0, 1] independently for each sequence in the spoken digits dataset. ::

        from sklearn.preprocessing import minmax_scale
        from sequentia.preprocessing import IndependentFunctionTransformer
        from sequentia.datasets import load_digits

        # Fetch MFCCs of spoken digits
        data = load_digits()

        # Create an independent min-max transform
        transform = IndependentFunctionTransformer(minmax_scale)

        # Apply the transform to the data
        Xt = transform.transform(data.X, data.lengths)
    """

    _parameter_constraints: dict = {
        "func": [callable, None],
        "inverse_func": [callable, None],
        "validate": ["boolean"],
        "check_inverse": ["boolean"],
        "kw_args": [dict, None],
        "inv_kw_args": [dict, None],
    }

    def __init__(
        self,
        func: Optional[Callable] = None,
        inverse_func: Optional[Callable] = None,
        *,
        validate: bool = False,
        check_inverse: bool = True,
        kw_args =None,
        inv_kw_args=None,
    ) -> IndependentFunctionTransformer:
        """Initializes the :class:`.IndependentFunctionTransformer`.

        :param func: The callable to use for the transformation.
            This will be passed the same arguments as transform, with args and kwargs forwarded.
            If ``None``, then ``func`` will be the identity function.

        :param inverse_func: The callable to use for the inverse transformation.
            This will be passed the same arguments as inverse transform, with args and kwargs forwarded.
            If ``None``, then ``inverse_func`` will be the identity function.

        :param validate: Indicates whether the input ``X`` array should be checked before calling ``func``.

            - If ``False``, there is no input validation.
            - If ``True``, then ``X`` will be converted to a 2-dimensional NumPy array.
              If the conversion is not possible an exception is raised.

        :param check_inverse: Whether to check that or ``func`` followed by ``inverse_func`` leads to the original inputs.
            It can be used for a sanity check, raising a warning when the condition is not fulfilled.

        :param kw_args: Dictionary of additional keyword arguments to pass to ``func``.

        :param inv_kw_args: Dictionary of additional keyword arguments to pass to ``inverse_func``.
        """
        self.func = func
        self.inverse_func = inverse_func
        self.validate = validate
        self.check_inverse = check_inverse
        self.kw_args: Optional[Dict[str, Any]] = kw_args
        self.inv_kw_args: Optional[Dict[str, Any]] = inv_kw_args

    def _check_input(self, X, lengths, *, reset):
        data = _BaseSequenceValidator(X=X, lengths=lengths)
        X, lengths = data.X, data.lengths
        if self.validate:
            X = self._validate_data(X, reset=reset)
            return X, lengths
        elif reset:
            # Set feature_names_in_ and n_features_in_ even if validate=False
            # We run this only when reset==True to store the attributes but not
            # validate them, because validate=False
            self._check_n_features(X, reset=reset)
            self._check_feature_names(X, reset=reset)
        return X, lengths

    def _check_inverse_transform(self, X, lengths):
        """Check that func and inverse_func are the inverse."""
        idx_selected = slice(None, None, max(1, X.shape[0] // 100))
        X_round_trip = self.inverse_transform(self.transform(X[idx_selected], lengths), lengths)

        if not np.issubdtype(X.dtype, np.number):
            raise ValueError(
                "'check_inverse' is only supported when all the elements in `X` are numerical."
            )

        if not _allclose_dense_sparse(X[idx_selected], X_round_trip):
            warnings.warn(
                "The provided functions are not strictly"
                " inverse of each other. If you are sure you"
                " want to proceed regardless, set"
                " 'check_inverse=False'.",
                UserWarning,
            )

    def fit(
        self,
        X: Array,
        lengths: Optional[Array] = None
    ) -> IndependentFunctionTransformer:
        """Fits the transformer to ``X``.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :return: The fitted transformer.
        """
        self._validate_params()
        X, lengths = self._check_input(X, lengths, reset=True)
        if self.check_inverse and not (self.func is None or self.inverse_func is None):
            self._check_inverse_transform(X, lengths)
        return self

    def transform(
        self,
        X: Array,
        lengths: Optional[Array] = None
    ) -> Array:
        """Applies the transformation to ``X``, producing a transformed version of ``X``.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :return: The transformed array.
        """
        X, lengths = self._check_input(X, lengths, reset=False)
        return self._transform(X, lengths, func=self.func, kw_args=self.kw_args)

    def inverse_transform(
        self,
        X: Array,
        lengths: Optional[Array] = None
    ) -> Array:
        """Applies the inverse transformation to ``X``.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :return: The inverse transformed array.
        """
        data = _BaseSequenceValidator(X=X, lengths=lengths)
        X, lengths = data.X, data.lengths
        if self.validate:
            X = check_array(X, accept_sparse=False)
        return self._transform(X, lengths, func=self.inverse_func, kw_args=self.inv_kw_args)

    def _transform(self, X, lengths, func=None, kw_args=None):
        if func is None:
            return X
        apply = lambda x: func(x, **(kw_args if kw_args else {}))
        idxs = SequentialDataset._get_idxs(lengths)
        return np.vstack([apply(x) for x in SequentialDataset._iter_X(X, idxs)])

    def __sklearn_is_fitted__(self):
        """Return True since FunctionTransfomer is stateless."""
        return True

    def _more_tags(self):
        return {"no_validation": not self.validate, "stateless": True}


def mean_filter(x: Array, k: PositiveInt = 3) -> Array:
    """Applies a mean filter of size ``k`` independently to each feature of the sequence,
    retaining the original input shape by using appropriate padding.

    This is implemented as a 1D convolution with a kernel of size ``k`` and values ``1 / k``.

    :param x: Univariate or multivariate observation sequence.

        - Should be a single 1D or 2D array.
        - Should have length as the 1st dimension and features as the 2nd dimension.

    :param k: Width of the filter.

    :return: The filtered array.

    Examples
    --------
    Applying a :func:`mean_filter` to a single sequence
    and multiple sequences (independently via :class:`IndependentFunctionTransformer`) from the spoken digits dataset. ::

        from sequentia.preprocessing import IndependentFunctionTransformer, mean_filter
        from sequentia.datasets import load_digits

        # Fetch MFCCs of spoken digits
        data = load_digits()

        # Apply the mean filter to the first sequence
        x, _ = data[0]
        xt = mean_filter(x, k=5)

        # Create an independent mean filter transform
        transform = IndependentFunctionTransformer(mean_filter, kw_args={"k": 5})

        # Apply the transform to all sequences
        Xt = transform.transform(data.X, data.lengths)
    """
    data = _BaseSequenceValidator(X=x)
    return convolve(data.X, np.ones((k, 1)) / k, mode="same")


def median_filter(x: Array, k: PositiveInt = 3) -> Array:
    """Applies a median filter of size ``k`` independently to each feature of the sequence,
    retaining the original input shape by using appropriate padding.

    :param x: Univariate or multivariate observation sequence.

        - Should be a single 1D or 2D array.
        - Should have length as the 1st dimension and features as the 2nd dimension.

    :param k: Width of the filter.

    :return: The filtered array.

    Examples
    --------
    Applying a :func:`median_filter` to a single sequence
    and multiple sequences (independently via :class:`IndependentFunctionTransformer`) from the spoken digits dataset. ::

        from sequentia.preprocessing import IndependentFunctionTransformer, median_filter
        from sequentia.datasets import load_digits

        # Fetch MFCCs of spoken digits
        data = load_digits()

        # Apply the median filter to the first sequence
        x, _ = data[0]
        xt = median_filter(x, k=5)

        # Create an independent median filter transform
        transform = IndependentFunctionTransformer(median_filter, kw_args={"k": 5})

        # Apply the transform to all sequences
        Xt = transform.transform(data.X, data.lengths)
    """
    data = _BaseSequenceValidator(X=x)
    return medfilt2d(data.X, kernel_size=(k, 1))



import pytest

from sequentia.datasets import load_digits

from ...support.assertions import assert_equal


@pytest.mark.parametrize('digits', [list(range(10)), [2, 5]])
def test_digits(digits):
    data = load_digits(digits=digits)

    assert len(data) == 300 * len(digits)
    assert_equal(data.classes, digits)
    assert set(data.y) == set(digits)

    for _, lengths, c in data.iter_by_class():
        assert len(lengths) == 300



import pytest
from typing import List
from dataclasses import dataclass
from operator import itemgetter

from pydantic import BaseModel, ValidationError
from sklearn.exceptions import NotFittedError

from sequentia.models import KNNClassifier
from sequentia.utils.decorators import _validate_params, _requires_fit, _override_params

def test_validate_params():
    class Validator(BaseModel):
        param: List[int]

    @_validate_params(using=Validator)
    def to_validate(*, param):
        pass

    with pytest.raises(ValidationError):
        to_validate(param=None)


def test_requires_fit():
    class Model:
        def fit(self):
            self.fitted_ = True

        @_requires_fit
        def predict(self):
            pass

    model = Model()

    with pytest.raises(NotFittedError):
        model.predict()

    model.fit()
    model.predict()


@pytest.mark.parametrize('temporary', [True, False])
@pytest.mark.parametrize('error', [True, False])
def test_override_params(temporary, error):
    @dataclass
    class Model:
        b: int = 1
        c: int = 2

        @_override_params(['b', 'c'], temporary=temporary)
        def evaluate(self, a, **kwargs):
            for param in ('b', 'c'):
                if param in kwargs:
                    assert getattr(self, param) == kwargs[param]
            if error:
                raise ValueError()

    model = Model()

    try:
        model.evaluate(a=0)
    except ValueError:
        pass

    assert model.b == 1 and model.c == 2

    if temporary:
        try:
            model.evaluate(a=0, b=2, c=1)
        except ValueError:
            pass

        assert model.b == 1 and model.c == 2
    else:
        try:
            model.evaluate(a=0, b=2, c=1)
        except ValueError:
            pass

        assert model.b == 2 and model.c == 1



"""
Pipeline is an adapted version of Pipeline from the sklearn.pipeline module,
and largely relies on its source code.

Below is the original license from Scikit-Learn, copied on 31st December 2022
from https://github.com/scikit-learn/scikit-learn/blob/main/COPYING.

---

BSD 3-Clause License

Copyright (c) 2007-2022 The scikit-learn developers.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

from __future__ import annotations

from typing import Optional, Any, Union, List, Tuple

import sklearn.pipeline
import sklearn.base
from joblib import Memory

from sklearn.pipeline import _final_estimator_has
from sklearn.utils.metaestimators import available_if
from sklearn.utils import _print_elapsed_time
from sklearn.utils.validation import check_memory
from sklearn.base import BaseEstimator, ClassifierMixin, TransformerMixin, clone

from sequentia.preprocessing.base import Transform
from sequentia.utils.validation import Array

__all__ = ["Pipeline"]


class Pipeline(sklearn.pipeline.Pipeline):
    """Pipeline of transforms with a final estimator.

    Sequentially apply a list of transforms and a final estimator.
    Intermediate steps of the pipeline must be 'transforms', that is, they
    must implement ``fit`` and ``transform`` methods.
    The final estimator only needs to implement ``fit``.
    The transformers in the pipeline can be cached using ``memory`` argument.
    The purpose of the pipeline is to assemble several steps that can be
    cross-validated together while setting different parameters. For this, it
    enables setting parameters of the various steps using their names and the
    parameter name separated by a ``__``. A step's estimator may be replaced
    entirely by setting the parameter with its name to another estimator,
    or a transformer removed by setting it to ``'passthrough'`` or ``None``.

    See Also
    --------
    :class:`sklearn.pipeline.Pipeline`:
        :class:`.Pipeline` is based on :class:`sklearn.pipeline.Pipeline`,
        but adapted to accept and work with sequences.

        Read more in the :ref:`User Guide <pipeline>`.

    Examples
    --------
    Creating a :class:`.Pipeline` consisting of two transforms and a :class:`.KNNClassifier`,
    and fitting it to sequences in the spoken digits dataset. ::

        from sequentia.models import KNNClassifier
        from sequentia.preprocessing import IndependentFunctionTransformer
        from sequentia.pipeline import Pipeline
        from sequentia.datasets import load_digits

        from sklearn.preprocessing import scale
        from sklearn.decomposition import PCA

        # Fetch MFCCs of spoken digits
        digits = load_digits()
        train, test = digits.split(test_size=0.2)

        # Create a pipeline with two transforms and a classifier
        pipeline = Pipeline([
            ('standardize', IndependentFunctionTransformer(scale)),
            ('pca', PCA(n_components=5)),
            ('clf', KNNClassifier(k=1))
        ])

        # Fit the pipeline transforms and classifier to training data
        pipeline.fit(train.X, train.lengths)

        # Apply the transforms to training sequences and make predictions
        y_train_pred = pipeline.predict(train.X, train.y, train.lengths)

        # Calculate accuracy on test data
        acc = pipeline.score(test.X, test.y, test.lengths)
    """

    def __init__(
        self,
        steps: List[Tuple[str, BaseEstimator]],
        *,
        memory: Optional[Union[str, Memory]] = None,
        verbose: bool = False
    ) -> Pipeline:
        """Initializes the :class:`.Pipeline`.

        :param steps: Collection of transforms implementing ``fit``/``transform`` that are chained,
            with the last object being an estimator.

        :param memory: Used to cache the fitted transformers of the pipeline. By default,
            no caching is performed. If a string is given, it is the path to
            the caching directory. Enabling caching triggers a clone of
            the transformers before fitting. Therefore, the transformer
            instance given to the pipeline cannot be inspected
            directly. Use the attribute ``named_steps`` or ``steps`` to
            inspect estimators within the pipeline. Caching the
            transformers is advantageous when fitting is time consuming.

        :param verbose: If ``True``, the time elapsed while fitting each step will be printed as it
            is completed.
        """
        super().__init__(steps, memory=memory, verbose=verbose)


    def _can_transform(self):
        return self._final_estimator == "passthrough" or hasattr(
            self._final_estimator, "transform"
        )


    def _can_inverse_transform(self):
        return all(hasattr(t, "inverse_transform") for _, _, t in self._iter())


    def _fit(
        self,
        X: Array,
        lengths: Optional[Array] = None,
        **fit_params_steps
    ) -> Array:
        # shallow copy of steps - this should really be steps_
        self.steps = list(self.steps)
        self._validate_steps()
        # Setup the memory
        memory = check_memory(self.memory)

        fit_transform_one_cached = memory.cache(_fit_transform_one)

        for step_idx, name, transformer in self._iter(
            with_final=False, filter_passthrough=False
        ):
            if transformer is None or transformer == "passthrough":
                with _print_elapsed_time("Pipeline", self._log_message(step_idx)):
                    continue

            if hasattr(memory, "location") and memory.location is None:
                # we do not clone when caching is disabled to
                # preserve backward compatibility
                cloned_transformer = transformer
            else:
                cloned_transformer = clone(transformer)
            # Fit or load from cache the current transformer
            X, fitted_transformer = fit_transform_one_cached(
                cloned_transformer,
                X,
                lengths,
                None,
                message_clsname="Pipeline",
                message=self._log_message(step_idx),
                **fit_params_steps[name],
            )
            # Replace the transformer of the step with the fitted
            # transformer. This is necessary when loading the transformer
            # from the cache.
            self.steps[step_idx] = (name, fitted_transformer)
        return X


    def fit(
        self,
        X: Array,
        y: Optional[Array] = None,
        lengths: Optional[Array] = None,
        **fit_params
    ) -> Pipeline:
        """Fit the model.

        Fit all the transformers one after the other and transform the
        data. Finally, fit the transformed data using the final estimator.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param y: Outputs corresponding to sequence(s) provided in ``X``.
            Only required if the final estimator is a supervised model.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :param fit_params: Parameters passed to the ``fit`` method of each step,
            where each parameter name is prefixed such that parameter ``p`` for step ``s`` has key ``s__p``.

        :return: The fitted pipeline.
        """
        fit_params_steps = self._check_fit_params(**fit_params)
        Xt = self._fit(X, lengths, **fit_params_steps)

        with _print_elapsed_time("Pipeline", self._log_message(len(self.steps) - 1)):
            if self._final_estimator != "passthrough":
                fit_params_last_step = fit_params_steps[self.steps[-1][0]]
                if isinstance(self._final_estimator, ClassifierMixin):
                    self._final_estimator.fit(Xt, y, lengths, **fit_params_last_step)
                else:
                    self._final_estimator.fit(Xt, lengths, **fit_params_last_step)
        return self


    def fit_transform(
        self,
        X: Array,
        lengths: Optional[Array] = None,
        **fit_params
    ) -> Array:
        """Fit the model and transform with the final estimator.

        Fits all the transformers one after the other and transform the
        data. Then uses `fit_transform` on transformed data with the final
        estimator.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :param fit_params: Parameters passed to the ``fit`` method of each step,
            where each parameter name is prefixed such that parameter ``p`` for step ``s`` has key ``s__p``.

        :return: The transformed data.
        """
        fit_params_steps = self._check_fit_params(**fit_params)
        Xt = self._fit(X, lengths, **fit_params_steps)

        last_step = self._final_estimator
        with _print_elapsed_time("Pipeline", self._log_message(len(self.steps) - 1)):
            if last_step == "passthrough":
                return Xt
            fit_params_last_step = fit_params_steps[self.steps[-1][0]]

            if is_sequentia_transform := isinstance(last_step, Transform):
                fit_params_last_step["lengths"] = lengths

            if hasattr(last_step, "fit_transform"):
                return last_step.fit_transform(Xt, **fit_params_last_step)
            else:
                transform_params = {}
                if is_sequentia_transform:
                    transform_params["lengths"] = lengths
                getattr(last_step, "transform")
                return last_step.fit(Xt, **fit_params_last_step).transform(Xt, **transform_params)


    @available_if(_final_estimator_has("predict"))
    def predict(
        self,
        X: Array,
        lengths: Optional[Array] = None
    ) -> Array:
        """Transform the data, and apply `predict` with the final estimator.

        Call `transform` of each transformer in the pipeline. The transformed
        data are finally passed to the final estimator that calls `predict`
        method. Only valid if the final estimator implements `predict`.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :return: Output predictions.
        """
        Xt = X
        for _, name, transform in self._iter(with_final=False):
            transform_params = {}
            if isinstance(transform, Transform):
                transform_params["lengths"] = lengths
            Xt = transform.transform(Xt, **transform_params)
        return self.steps[-1][1].predict(Xt, lengths)


    @available_if(_final_estimator_has("fit_predict"))
    def fit_predict(
        self,
        X: Array,
        y: Array,
        lengths: Optional[Array] = None,
        **fit_params
    ) -> Array:
        """Transform the data, and apply `fit_predict` with the final estimator.

        Call `fit_transform` of each transformer in the pipeline. The
        transformed data are finally passed to the final estimator that calls
        `fit_predict` method. Only valid if the final estimator implements
        `fit_predict`.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param y: Outputs corresponding to sequence(s) provided in ``X``.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :param fit_params: Parameters passed to the ``fit`` method of each step,
            where each parameter name is prefixed such that parameter ``p`` for step ``s`` has key ``s__p``.

        :return: Output predictions.
        """
        fit_params_steps = self._check_fit_params(**fit_params)
        Xt = self._fit(X, lengths, **fit_params_steps)

        fit_params_last_step = fit_params_steps[self.steps[-1][0]]
        with _print_elapsed_time("Pipeline", self._log_message(len(self.steps) - 1)):
            y_pred = self.steps[-1][1].fit_predict(Xt, y, lengths, **fit_params_last_step)
        return y_pred


    @available_if(_final_estimator_has("predict_proba"))
    def predict_proba(
        self,
        X: Array,
        lengths: Optional[Array] = None
    ) -> Array:
        """Transform the data, and apply `predict_proba` with the final estimator.

        Call `transform` of each transformer in the pipeline. The transformed
        data are finally passed to the final estimator that calls
        `predict_proba` method. Only valid if the final estimator implements
        `predict_proba`.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :return: Output probabilities.
        """
        Xt = X
        for _, name, transform in self._iter(with_final=False):
            transform_params = {}
            if isinstance(transform, Transform):
                transform_params["lengths"] = lengths
            Xt = transform.transform(Xt, **transform_params)
        return self.steps[-1][1].predict_proba(Xt, lengths)


    @available_if(_can_transform)
    def transform(
        self,
        X: Array,
        lengths: Optional[Array] = None
    ) -> Array:
        """Transform the data, and apply `transform` with the final estimator.

        Call `transform` of each transformer in the pipeline. The transformed
        data are finally passed to the final estimator that calls
        `transform` method. Only valid if the final estimator
        implements `transform`.

        This also works where final estimator is `None` in which case all prior
        transformations are applied.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :return: The transformed data.
        """
        Xt = X
        for _, _, transform in self._iter():
            transform_params = {}
            if isinstance(transform, Transform):
                transform_params["lengths"] = lengths
            Xt = transform.transform(Xt, **transform_params)
        return Xt


    @available_if(_can_inverse_transform)
    def inverse_transform(
        self,
        X: Array,
        lengths: Optional[Array] = None
    ) -> Array:
        """Apply `inverse_transform` for each step in a reverse order.

        All estimators in the pipeline must support `inverse_transform`.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :return: The inverse transformed data.
        """
        reverse_iter = reversed(list(self._iter()))
        for _, _, transform in reverse_iter:
            transform_params = {}
            if isinstance(transform, Transform):
                transform_params["lengths"] = lengths
            X = transform.inverse_transform(X, **transform_params)
        return X


    @available_if(_final_estimator_has("score"))
    def score(
        self,
        X: Array,
        y: Optional[Array] = None,
        lengths: Optional[Array] = None,
        sample_weight: Optional[Any] = None
    ) -> float:
        """Transform the data, and apply `score` with the final estimator.

        Call `transform` of each transformer in the pipeline. The transformed
        data are finally passed to the final estimator that calls
        `score` method. Only valid if the final estimator implements `score`.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D or 2D array.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param y: Outputs corresponding to sequence(s) provided in ``X``.
            Must be provided if the final estimator is a model, i.e. not a transform.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :param sample_weight: If not ``None``, this argument is passed as ``sample_weight``
            keyword argument to the ``score`` method of the final estimator.

        :return: Result of calling `score` on the final estimator.
        """
        Xt = X
        for _, name, transform in self._iter(with_final=False):
            transform_params = {}
            if isinstance(transform, Transform):
                transform_params["lengths"] = lengths
            Xt = transform.transform(Xt, **transform_params)

        score_params = {}
        if sample_weight is not None:
            score_params["sample_weight"] = sample_weight

        last_step = self.steps[-1][1]
        if isinstance(last_step, TransformerMixin):
            return last_step.score(Xt, lengths, **score_params)
        else:
            return last_step.score(Xt, y, lengths, **score_params)


def _fit_transform_one(
    transformer,
    X,
    lengths,
    weight,
    message_clsname="",
    message=None,
    **fit_params
):
    with _print_elapsed_time(message_clsname, message):
        if is_sequentia_transformer := isinstance(transformer, Transform):
            fit_params["lengths"] = lengths
        if hasattr(transformer, "fit_transform"):
            res = transformer.fit_transform(X, **fit_params)
        else:
            transform_params = {}
            if is_sequentia_transformer:
                transform_params["lengths"] = lengths
            res = transformer.fit(X, **fit_params).transform(X, **transform_params)

    if weight is None:
        return res, transformer
    return res * weight, transformer


if __name__ == "__main__":
    import numpy as np

    from sequentia.models import KNNClassifier, HMMClassifier, GaussianMixtureHMM
    from sequentia.datasets import load_digits
    from sequentia.preprocessing import IndependentFunctionTransformer
    # try normal FunctionTransformer

    from sklearn.preprocessing import StandardScaler, scale
    from sklearn.decomposition import PCA

    random_state = np.random.RandomState(0)

    # digits = load_digits(digits=[0, 1])
    digits = load_digits()
    # subset, _ = digits.split(train_size=0.1, stratify=True, random_state=random_state)
    # train, test = subset.split(test_size=0.2)
    train, test = digits.split(test_size=0.2)

    pipeline = Pipeline([
        ('standardize', IndependentFunctionTransformer(scale)),
        ('pca', PCA(n_components=5, random_state=random_state)),
        ('clf', HMMClassifier(n_jobs=-1).add_models({
            k: GaussianMixtureHMM(random_state=random_state)
            for k in train.classes
        }))
        # ('clf', KNNClassifier(k=1, use_c=True, n_jobs=-1, random_state=random_state))
    ])

    # Xt = pipeline.fit_transform(*train.X_lengths)

    y_pred = pipeline.fit_predict(*train.X_y_lengths)

    breakpoint()

    # from scipy.signal import medfilt2d, convolve
    # ('median_filter', IndependentFunctionTransformer(medfilt2d, kw_args={"kernel_size": (5, 1)}))



from itertools import chain, combinations as combinations_

def combinations(string):
    return map(lambda params: ''.join(params), chain.from_iterable(combinations_(string, i) for i in range(1, len(string))))


from __future__ import annotations

from typing import Optional, Any

from sklearn.base import BaseEstimator, ClassifierMixin, RegressorMixin
from sklearn.metrics import accuracy_score, r2_score

from sequentia.utils.validation import Array
from sequentia.utils.decorators import _requires_fit


class _Classifier(BaseEstimator, ClassifierMixin):
    def fit(
        self,
        X: Array,
        y: Array[int],
        lengths: Optional[Array[int]] = None
    ) -> _Classifier:
        raise NotImplementedError


    def predict(
        self,
        X: Array,
        lengths: Optional[Array[int]] = None
    ) -> Array[int]:
        raise NotImplementedError


    def fit_predict(
        self,
        X: Array,
        y: Array[int],
        lengths: Optional[Array[int]] = None
    ) -> Array[int]:
        return self.fit(X, y, lengths).predict(X, lengths)


    def predict_proba(
        self,
        X: Array,
        lengths: Optional[Array[int]] = None
    ) -> Array[float]:
        raise NotImplementedError


    def predict_scores(
        self,
        X: Array,
        lengths: Optional[Array[int]] = None
    ) -> Array[float]:
        raise NotImplementedError


    @_requires_fit
    def score(
        self,
        X: Array,
        y: Array[int],
        lengths: Optional[Array[int]] = None,
        normalize: bool = True,
        sample_weight: Optional[Any] = None
    ) -> float:
        y_pred = self.predict(X, lengths)
        return accuracy_score(y, y_pred, normalize=normalize, sample_weight=sample_weight)


class _Regressor(BaseEstimator, RegressorMixin):
    def fit(
        self,
        X: Array[float],
        y: Array[float],
        lengths: Optional[Array[int]] = None
    ) -> _Regressor:
        raise NotImplementedError


    def predict(
        self,
        X: Array[float],
        lengths: Optional[Array[int]] = None
    ) -> Array[float]:
        raise NotImplementedError


    def fit_predict(
        self,
        X: Array[float],
        y: Array[float],
        lengths: Optional[Array[int]] = None
    ) -> Array[float]:
        return self.fit(X, y, lengths).predict(X, lengths)


    @_requires_fit
    def score(
        self,
        X: Array[float],
        y: Array[float],
        lengths: Optional[Array[int]] = None,
        sample_weight: Optional[Any] = None
    ) -> float:
        y_pred = self.predict(X, lengths)
        return r2_score(y, y_pred, sample_weight=sample_weight)



from __future__ import annotations

import joblib
import pathlib
from types import SimpleNamespace
from typing import Optional, Union, Dict, Literal, IO
from joblib import Parallel, delayed

import numpy as np
from pydantic import NegativeInt, PositiveInt, confloat, validator, root_validator
from sklearn.utils.validation import NotFittedError

from sequentia.models.base import _Classifier
from sequentia.models.hmm.variants.base import _HMM
from sequentia.utils.data import SequentialDataset
from sequentia.utils.multiprocessing import _effective_n_jobs
from sequentia.utils.decorators import _validate_params, _override_params, _requires_fit
from sequentia.utils.validation import (
    _check_classes,
    _check_is_fitted,
    Array,
    _Validator,
)

__all__ = ['HMMClassifier']

_defaults = SimpleNamespace(
    prior=None,
    classes=None,
    n_jobs=1,
)


class _HMMClassifierValidator(_Validator):
    prior: Optional[Union[Literal["frequency"], Dict[int, confloat(ge=0, le=1)]]] = _defaults.prior
    classes: Optional[Array[int]] = _defaults.classes
    n_jobs: Union[NegativeInt, PositiveInt] = _defaults.n_jobs


    @validator('prior')
    def check_prior(cls, value):
        if isinstance(value, dict):
            if not np.isclose(sum(value.values()), 1):
                raise ValueError('Prior distribution must sum to one')
        return value


    @root_validator
    def check_prior_keys_with_classes(cls, values):
        if 'prior' in values and 'classes' in values:
            prior, classes = values['prior'], values['classes']
            if isinstance(prior, dict) and classes is not None:
                if set(prior.keys()) != set(classes):
                    raise ValueError(
                        'Provided classes are not consistent with the provided prior distribution - '
                        'ensure that every label in `classes` is present in `prior`'
                    )
        return values


class HMMClassifier(_Classifier):
    """A classifier consisting of HMMs, each trained independently to recognize sequences of a single class.

    The predicted class for a given observation sequence is the class represented by the HMM
    which produces the maximum posterior probability for the observation sequence.

    Examples
    --------
    Using a :class:`.HMMClassifier` (with :class:`.GaussianMixtureHMM` models) to classify spoken digits. ::

        import numpy as np
        from sequentia.datasets import load_digits
        from sequentia.models.hmm import GaussianMixtureHMM, HMMClassifier

        # Seed for reproducible pseudo-randomness
        random_state = np.random.RandomState(1)

        # Fetch MFCCs of spoken digits
        data = load_digits()
        train_data, test_data = data.split(test_size=0.2, random_state=random_state)

        # Create a HMMClassifier using a class frequency prior
        clf = HMMClassifier(prior='frequency')

        # Add an untrained HMM for each class
        for label in data.classes:
            model = GaussianMixtureHMM(random_state=random_state)
            clf.add_model(model, label)

        # Fit the HMMs by providing training observation sequences for all classes
        X_train, y_train, lengths_train = train_data.X_y_lengths
        clf.fit(X_train, y_train, lengths_train)

        # Predict classes for the test observation sequences
        X_test, lengths_test = test_data.X_lengths
        y_pred = clf.predict(X_test, lengths_test)

    As done in the above example, we can provide unfitted HMMs using :func:`add_model` or :func:`add_models`,
    then provide training observation sequences for all classes to :func:`fit`, which will automatically train each HMM on the appropriate subset of data.

    Alternatively, we may provide pre-fitted HMMs and call :func:`fit` with no arguments. ::

        # Create a HMMClassifier using a class frequency prior
        clf = HMMClassifier(prior='frequency')

       # Manually fit each HMM on its own subset of data
        for X_train, lengths_train, label for train_data.iter_by_class():
            model = GaussianMixtureHMM(random_state=random_state)
            model.fit(X_train, lengths_train)
            clf.add_model(model, label)

        # Fit the classifier
        clf.fit()
    """

    _defaults = _defaults


    @_validate_params(using=_HMMClassifierValidator)
    def __init__(
        self,
        *,
        prior: Optional[Union[Literal["frequency"], dict]] = _defaults.prior,
        classes: Optional[Array[int]] = _defaults.classes,
        n_jobs: Union[NegativeInt, PositiveInt] = _defaults.n_jobs,
    ) -> HMMClassifier:
        """Initializes a :class:`.HMMClassifier`.

        :param prior: Type of prior probability to assign to each HMM.

            - If ``None``, a uniform prior will be used, making each HMM equally likely.
            - If ``"frequency"``, the prior probability of each HMM is equal to the fraction of total observation sequences that the HMM was fitted with.
            - If a ``dict``, custom prior probabilities can be assigned to each HMM.
              The keys should be the label of the class represented by the HMM, and the value should be the prior probability for the HMM.

        :param classes: Set of possible class labels.

            - If not provided, these will be determined from the training data labels.
            - If provided, output from methods such as :func:`predict_proba` and :func:`predict_scores`
              will follow the ordering of the class labels provided here.

        :param n_jobs: Maximum number of concurrently running workers.

            - If 1, no parallelism is used at all (useful for debugging).
            - If -1, all CPUs are used.
            - If < -1, ``(n_cpus + 1 + n_jobs)`` are used — e.g. ``n_jobs=-2`` uses all but one.
        """
        #: Type of prior probability to assign to each HMM.
        self.prior = prior
        #: Set of possible class labels.
        self.classes = classes
        #: Maximum number of concurrently running workers.
        self.n_jobs = n_jobs
        #: HMMs constituting the :class:`.HMMClassifier`.
        self.models = {}


    def add_model(
        self,
        model: _HMM,
        label: int
    ) -> HMMClassifier:
        """Adds a single HMM to the classifier.

        :param model: HMM to add to the classifier.
        :param label: Class represented by the HMM.

        :note: All models added to the classifier must be of the same type — either :class:`.GaussianMixtureHMM` or :class:`.CategoricalHMM`.

        :return: The classifier.
        """
        if not isinstance(model, _HMM):
            raise TypeError('Expected `model` argument to be a type of HMM')
        if len(self.models) > 0:
            if type(model) != type(list(self.models.values())[-1]):
                raise TypeError(
                    f'Model of type {type(model).__name__} must be the same as the models already provided '
                    f'to this {type(self).__name__} instance'
                )
        self.models[int(label)] = model
        return self


    def add_models(
        self,
        models: Dict[int, _HMM]
    ) -> HMMClassifier:
        """Adds HMMs to the classifier.

        :param models: HMMs to add to the classifier. The key for each HMM should be the label of the class represented by the HMM.

        :note: All models added to the classifier must be of the same type — either :class:`.GaussianMixtureHMM` or :class:`.CategoricalHMM`.

        :return: The classifier.
        """
        if not all(isinstance(model, _HMM) for model in models.values()):
            raise TypeError('Expected all provided `models` to be a type of HMM')
        for label, model in models.items():
            self.add_model(model, label)
        return self


    def fit(
        self,
        X: Optional[Array] = None,
        y: Optional[Array[int]] = None,
        lengths: Optional[Array[int]] = None
    ) -> HMMClassifier:
        """Fits the HMMs to the sequence(s) in ``X``.

        - If fitted models were provided with :func:`add_model` or :func:`add_models`, no arguments should be passed to :func:`fit`.
        - If unfitted models were provided with :func:`add_model` or :func:`add_models`, training data ``X``, ``y`` and ``lengths`` must be provided to :func:`fit`.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D array if :class:`.CategoricalHMM` is being used, or either a 1D or 2D array if :class:`.GaussianMixtureHMM` is being used.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param y: Classes corresponding to sequence(s) provided in ``X``.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :return: The fitted classifier.
        """
        if X is None or y is None:
            if len(self.models) == 0:
                raise RuntimeError(
                    f'Fitted models must be provided to this {type(self).__name__} instance if no training data is provided - '
                    'use add_model() to add fitted models to the classifier object'
                )

            for label, model in self.models.items():
                if not _check_is_fitted(model, return_=True):
                    raise NotFittedError(
                        f'The model corresponding to label {label} must be pre-fitted if '
                        f'no training data is provided to this {type(self).__name__} instance'
                    )

            if self.classes is not None:
                # Same logic as _check_classes()
                classes_np = np.array(self.classes).flatten()
                if not np.issubdtype(classes_np.dtype, np.integer):
                    raise TypeError(f'Expected classes to be integers')
                _, idx = np.unique(classes_np, return_index=True)
                self.classes_ = classes_np[np.sort(idx)]
            else:
                # Fetch classes from provided models
                self.classes_ = np.array(list(self.models.keys()))
        else:
            self.classes_ = _check_classes(Array[int].validate_type(y), self.classes)

        # Check that each label has a HMM (and vice versa)
        if set(self.models.keys()) != set(self.classes_):
            raise ValueError(
                'Classes in the dataset are not consistent with the added models - '
                'ensure that every added model corresponds to a class in the dataset'
            )

        if X is not None and y is not None:
            # Iterate through the dataset by class and fit the corresponding model
            data = self._sequence_classifier_validator(X=X, y=y, lengths=lengths)
            dataset = SequentialDataset(data.X, data.y, data.lengths, self.classes_)
            for X_c, lengths_c, c in dataset.iter_by_class():
                self.models[c].fit(X_c, lengths_c)

        # Set class priors
        if self.prior is None:
            self.prior_ = {c:1/len(self.classes_) for c, _ in self.models.items()}
        elif isinstance(self.prior, str):
            if self.prior == "frequency":
                total_seqs = sum(model.n_seqs_ for model in self.models.values())
                self.prior_ = {c:model.n_seqs_/total_seqs for c, model in self.models.items()}
        elif isinstance(self.prior, dict):
            if set(self.prior.keys()) != set(self.classes_):
                raise ValueError(
                    'Classes in the dataset are not consistent with the classes in `prior` - '
                    'ensure that every provided class prior corresponds to a class in the dataset'
                )
            self.prior_ = self.prior

        return self


    @_requires_fit
    def predict(
        self,
        X: Array,
        lengths: Optional[Array[int]] = None
    ) -> Array[int]:
        """Predicts classes for the sequence(s) in ``X``.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D array if :class:`.CategoricalHMM` is being used, or either a 1D or 2D array if :class:`.GaussianMixtureHMM` is being used.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :note: This method requires a trained classifier — see :func:`fit`.

        :return: Class predictions.
        """
        scores = self.predict_scores(X, lengths)
        max_score_idxs = scores.argmax(axis=1)
        return self.classes_[max_score_idxs]


    def fit_predict(
        self,
        X: Array,
        y: Array[int],
        lengths: Optional[Array[int]] = None
    ) -> Array[int]:
        """Fits the classifier to the sequence(s) in ``X`` and predicts classes for ``X``.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D array if :class:`.CategoricalHMM` is being used, or either a 1D or 2D array if :class:`.GaussianMixtureHMM` is being used.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param y: Classes corresponding to sequence(s) provided in ``X``.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :return: Class predictions.
        """
        return super().fit_predict(X, y, lengths)


    @_requires_fit
    def predict_proba(
        self,
        X: Array,
        lengths: Optional[Array[int]] = None
    ) -> Array[confloat(ge=0, le=1)]:
        """Predicts class probabilities for the sequence(s) in ``X``.

        Probabilities are calculated as the posterior probability of each HMM generating the sequence.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D array if :class:`.CategoricalHMM` is being used, or either a 1D or 2D array if :class:`.GaussianMixtureHMM` is being used.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :note: This method requires a trained classifier — see :func:`fit`.

        :return: Class membership probabilities.
        """
        proba = self.predict_scores(X, lengths)
        proba -= proba.max(axis=1, keepdims=True)
        proba = np.exp(proba)
        proba /= proba.sum(axis=1, keepdims=True)
        return proba


    @_requires_fit
    def predict_scores(
        self,
        X: Array,
        lengths: Optional[Array[int]] = None
    ) -> Array[float]:
        """Predicts class scores for the sequence(s) in ``X``.

        Scores are calculated as the log posterior probability of each HMM generating the sequence.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D array if :class:`.CategoricalHMM` is being used, or either a 1D or 2D array if :class:`.GaussianMixtureHMM` is being used.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :note: This method requires a trained classifier — see :func:`fit`.

        :return: Class scores.
        """
        data = self._base_sequence_validator(X=X, lengths=lengths)
        n_jobs = _effective_n_jobs(self.n_jobs, data.lengths)
        chunk_idxs = np.array_split(SequentialDataset._get_idxs(data.lengths), n_jobs)
        return np.concatenate(
            Parallel(n_jobs=n_jobs, max_nbytes=None)(
                delayed(self._compute_scores_chunk)(idxs, data.X)
                for idxs in chunk_idxs
            )
        )


    @_requires_fit
    def score(
        self,
        X: Array,
        y: Array[int],
        lengths: Optional[Array[int]],
        normalize: bool = True,
        sample_weight: Optional[Array] = None,
    ) -> float:
        """Calculates accuracy for the sequence(s) in ``X``.

        :param X: Univariate or multivariate observation sequence(s).

            - Should be a single 1D array if :class:`.CategoricalHMM` is being used, or either a 1D or 2D array if :class:`.GaussianMixtureHMM` is being used.
            - Should have length as the 1st dimension and features as the 2nd dimension.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param y: Classes corresponding to the observation sequence(s) in ``X``.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :param normalize: See :func:`sklearn:sklearn.metrics.accuracy_score`.

        :param sample_weight: See :func:`sklearn:sklearn.metrics.accuracy_score`.

        :note: This method requires a trained classifier — see :func:`fit`.

        :return: Classification accuracy.
        """
        return super().score(X, y, lengths, normalize, sample_weight)


    @_validate_params(using=_HMMClassifierValidator)
    @_override_params(_HMMClassifierValidator.fields(), temporary=False)
    def set_params(self, **kwargs) -> HMMClassifier:
        return self


    def _compute_scores_chunk(self, idxs, X):
        scores = np.zeros((len(idxs), len(self.classes_)))
        for i, x in enumerate(SequentialDataset._iter_X(X, idxs)):
            scores[i] = self._compute_log_posterior(x)
        return scores


    def _compute_log_posterior(self, x):
        log_posterior = np.full(len(self.classes_), -np.inf)
        for i, k in enumerate(self.classes_):
            model = self.models[k]
            log_prior = np.log(self.prior_[k])
            log_likelihood = model._score(x)
            log_posterior[i] = log_prior + log_likelihood
        return log_posterior


    def _base_sequence_validator(self, **kwargs):
        model = self.models[0]
        return model._base_sequence_validator(**kwargs)


    def _sequence_classifier_validator(self, **kwargs):
        model = self.models[0]
        return model._sequence_classifier_validator(**kwargs)


    @_requires_fit
    def save(self, path: Union[str, pathlib.Path, IO]):
        """Serializes and saves a fitted HMM classifier.

        :param path: Location to save the serialized classifier.

        :note: This method requires a trained classifier — see :func:`fit`.

        See Also
        --------
        load:
            Loads and deserializes a fitted HMM classifier.
        """
        # Fetch main parameters and fitted values
        state = {
            'params': self.get_params(),
            'models': self.models,
            'fitted': {k:v for k, v in self.__dict__.items() if k.endswith('_')}
        }

        # Serialize model
        joblib.dump(state, path)


    @classmethod
    def load(cls, path: Union[str, pathlib.Path, IO]) -> HMMClassifier:
        """Loads and deserializes a fitted HMM classifier.

        :param path: Location to load the serialized classifier from.

        :return: Fitted HMM classifier.

        See Also
        --------
        save:
            Serializes and saves a fitted HMM classifier.
        """
        state = joblib.load(path)

        # Set main parameters
        model = cls(**state['params'])
        model.models = state['models']

        # Set fitted values
        for k, v in state['fitted'].items():
            setattr(model, k, v)

        # Return deserialized model
        return model



from __future__ import annotations

from copy import deepcopy
from types import SimpleNamespace
from typing import Optional, Union, Dict, Any, Literal
from pydantic import NonNegativeInt, PositiveInt

import numpy as np
import hmmlearn.hmm
from sklearn.utils import check_random_state

from sequentia.models.hmm.topologies import _topologies
from sequentia.models.hmm.variants.base import _HMM, _HMMValidator
from sequentia.utils.decorators import _validate_params, _requires_fit
from sequentia.utils.validation import (
    Array,
    _BaseUnivariateCategoricalSequenceValidator,
    _SingleUnivariateCategoricalSequenceValidator,
    _UnivariateCategoricalSequenceClassifierValidator,
)

__all__ = ['CategoricalHMM']

_defaults = SimpleNamespace(
    **{
        **_HMM._defaults.__dict__,
        "hmmlearn_kwargs": dict(
            init_params="ste",
            params="ste",
        )
    }
)

class CategoricalHMM(_HMM):
    """A hidden Markov model with univariate categorical emissions.

    Examples
    --------
    Using a :class:`.CategoricalHMM` to learn how to recognize DNA sequences from the synthetase gene family.

    See :func:`.load_gene_families` for more information on the sample dataset used in this example.

    ::

        import numpy as np
        from sequentia.datasets import load_gene_families
        from sequentia.models.hmm import CategoricalHMM

        # Seed for reproducible pseudo-randomness
        random_state = np.random.RandomState(1)

        # Fetch DNA sequences for the synthetase gene family (no. 4)
        data, enc = load_gene_families(families=[4])
        train_data, test_data = data.split(test_size=0.2, random_state=random_state)

        # Create and train a CategoricalHMM to recognize the synthetase DNA sequences
        model = CategoricalHMM(random_state=random_state)
        X_train, lengths_train = train_data.X_lengths
        model.fit(X_train, lengths_train)

        # Calculate the log-likelihood of the first test sample being generated by this model
        x, y = test_data[0]
        model.score(x)
    """

    _base_sequence_validator = _BaseUnivariateCategoricalSequenceValidator
    _single_sequence_validator = _SingleUnivariateCategoricalSequenceValidator
    _sequence_classifier_validator = _UnivariateCategoricalSequenceClassifierValidator
    _defaults = _defaults

    def __init__(
        self,
        *,
        n_states: PositiveInt = _defaults.n_states,
        topology: Optional[Literal["ergodic", "left-right", "linear"]] = _defaults.topology,
        random_state: Optional[Union[NonNegativeInt, np.random.RandomState]] = _defaults.random_state,
        hmmlearn_kwargs: Dict[str, Any] = deepcopy(_defaults.hmmlearn_kwargs)
    ) -> CategoricalHMM:
        """Initializes the :class:`.CategoricalHMM`.

        :param n_states: Number of states in the Markov chain.
        :param topology: Transition topology of the Markov chain — see :ref:`topologies`.

          - If ``None``, behaves the same as ``'ergodic'`` but with `hmmlearn <https://hmmlearn.readthedocs.io/en/latest/>`__ initialization.

        :param random_state: Seed or :class:`numpy:numpy.random.RandomState` object for reproducible pseudo-randomness.
        :param hmmlearn_kwargs: Additional key-word arguments provided to the `hmmlearn <https://hmmlearn.readthedocs.io/en/latest/>`__ HMM constructor.
        """
        super().__init__(n_states, topology, random_state, hmmlearn_kwargs)

    def fit(
        self,
        X: Array[int],
        lengths: Optional[Array[int]] = None
    ) -> CategoricalHMM:
        """Fits the HMM to the sequences in ``X``, using the Baum—Welch algorithm.

        :param X: Univariate observation sequence(s).

            - Should be a single 1D array.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :return: The fitted HMM.
        """
        data = self._base_sequence_validator(X=X, lengths=lengths)
        self.random_state_ = check_random_state(self.random_state)
        if self.topology is None:
            self.topology_ = None
        else:
            self.topology_ = _topologies[self.topology](self.n_states, self.random_state_)
        self._check_init_params()

        kwargs = deepcopy(self.hmmlearn_kwargs)
        kwargs['init_params'] = ''.join(set(kwargs['init_params']) - self._skip_init_params)
        kwargs['params'] = ''.join(set(kwargs['params']) - self._skip_params)

        self.model = hmmlearn.hmm.CategoricalHMM(
            n_components=self.n_states,
            random_state=self.random_state_,
            **kwargs
        )

        for attr in ('startprob', 'transmat', 'emissionprob'):
            if hasattr(self, f'_{attr}'):
                setattr(self.model, f'{attr}_', getattr(self, f'_{attr}'))

        self.model.fit(data.X, lengths=data.lengths)
        self.n_seqs_ = len(data.lengths)

        return self

    @_requires_fit
    def score(
        self,
        x: Array[int],
    ) -> float:
        """Calculates the log-likelihood of the HMM generating a single observation sequence.

        :param x: Univariate observation sequence.

            - Should be a single 1D array.

        :note: This method requires a trained model — see :func:`fit`.

        :return: The log-likelihood.
        """
        return super().score(x)

    @_requires_fit
    def n_params(self) -> NonNegativeInt:
        n_params = super().n_params()
        if 'e' not in self._skip_params:
            n_params += self.model.emissionprob_.size
        return n_params

    @_requires_fit
    def bic(
        self,
        X: Array[int],
        lengths: Optional[Array[int]] = None
    ) -> float:
        """The Bayesian information criterion of the model, evaluated with the maximum likelihood of ``X``.

        :param X: Univariate observation sequence(s).

            - Should be a single 1D array.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :note: This method requires a trained model — see :func:`fit`.

        :return: The Bayesian information criterion.
        """
        return super().bic(X, lengths)

    @_requires_fit
    def aic(
        self,
        X: Array[int],
        lengths: Optional[Array[int]] = None
    ) -> float:
        """The Akaike information criterion of the model, evaluated with the maximum likelihood of ``X``.

        :param X: Univariate observation sequence(s).

            - Should be a single 1D array.
            - Should be a concatenated sequence if multiple sequences are provided,
              with respective sequence lengths being provided in the ``lengths`` argument for decoding the original sequences.

        :param lengths: Lengths of the observation sequence(s) provided in ``X``.

            - If ``None``, then ``X`` is assumed to be a single observation sequence.
            - ``len(X)`` should be equal to ``sum(lengths)``.

        :note: This method requires a trained model — see :func:`fit`.

        :return: The Akaike information criterion.
        """
        return super().aic(X, lengths)

    def set_state_emissions(self, values: Array[float]):
        """Sets the state emission distribution of the HMM's emission model.

        If this method is **not** called, emission probabilities will be initialized by `hmmlearn <https://hmmlearn.readthedocs.io/en/latest/>`__.

        :param values: Array of emission probabilities.

        :note: If used, this method should normally be called before :func:`fit`.
        """
        self._emissionprob = Array[float].validate_type(values)
        self._skip_init_params |= set('e')

    def freeze(
        self,
        params: str = deepcopy(_defaults.hmmlearn_kwargs["params"]),
    ):
        """Freezes the trainable parameters of the HMM, preventing them from being updated during the Baum—Welch algorithm.

        :param params: A string specifying which parameters to freeze. Can contain a combination of:

            - ``'s'`` for initial state probabilities,
            - ``'t'`` for transition probabilities,
            - ``'e'`` for emission probailities.

        :note: If used, this method should normally be called before :func:`fit`.

        See Also
        --------
        unfreeze:
            Unfreezes the trainable parameters of the HMM, allowing them to be updated during the Baum—Welch algorithm.
        """
        super().freeze(params)

    def unfreeze(
        self,
        params: str = deepcopy(_defaults.hmmlearn_kwargs["params"]),
    ):
        """Unfreezes the trainable parameters of the HMM, allowing them to be updated during the Baum—Welch algorithm.

        :param params: A string specifying which parameters to unfreeze. Can contain a combination of:

            - ``'s'`` for initial state probabilities,
            - ``'t'`` for transition probabilities,
            - ``'e'`` for emission probailities.

        See Also
        --------
        freeze:
            Freezes the trainable parameters of the HMM, preventing them from being updated during the Baum—Welch algorithm.
        """
        super().unfreeze(params)

class _CategoricalHMMValidator(_HMMValidator):
    hmmlearn_kwargs: Dict[str, Any] = deepcopy(_defaults.hmmlearn_kwargs)

    _class = CategoricalHMM

CategoricalHMM.__init__ = _validate_params(using=_CategoricalHMMValidator)(CategoricalHMM.__init__)



import pytest

from sequentia.datasets import load_gene_families

from ...support.assertions import assert_equal

counts = {
    0: 531,
    1: 534,
    2: 349,
    3: 672,
    4: 711,
    5: 240,
    6: 1343
}


@pytest.mark.parametrize('families', [list(range(7)), [2, 5]])
def test_gene_families(families):
    data, enc = load_gene_families(families=families)

    assert set(enc.classes_) == {'A', 'C', 'G', 'N', 'T'}

    assert_equal(data.classes, families)
    assert set(data.y) == set(families)

    for family in families:
        assert (data.y == family).sum() == counts[family]



from .datasets import *
from .models import *
from .preprocessing import *
from .pipeline import *
from .utils import *

__name__ = "sequentia"
__version__ = "1.1.0"
__author__ = "Edwin Onuonga"
__email__ = "ed@eonu.net"
__copyright__ = f"2019-2023, {__author__}"
